#include "Converter.h"

#include "Class.h"
#include "Data.h"
#include "Bitmap.h"
#include "DevBitmap.h"
#include "ETString.h"
#include "ColorMapper.h"
#include "TypeMatcher.h"
#include "ETMath.h"
#include "Error.h"
#include "Ink.h"

#define COLORS 256
static const Symbol cDocTypeGIF("GIF");

//---- GIFMatcher --------------------------------------------------------------

class GIFMatcher: public TypeMatcher {
public:
    GIFMatcher(const Symbol &t) : TypeMatcher(t, "gif", 0, FALSE)
	{ }
    bool MatchContents(Data*, char *buf, int len);
    bool MatchExtension(Data*, char *e)
	{ return StrCmp((byte *)e, (byte *)"gif", -1, upcase) == 0; }
    Bitmap *GetBitmap()
	{ return gImageIcon; }
};

bool GIFMatcher::MatchContents(Data*, char *b, int)
{
    return strncmp(b, "GIF87a", 6) == 0 || strncmp(b, "GIF89a", 6) == 0;
}


//---- GIFConverter ------------------------------------------------------------

class GIFConverter : public Converter {
public:
    MetaDef(GIFConverter);
    GIFConverter()
	{ }
    ~GIFConverter();
    char *AsString()
	{ return "Graphics Interchange Format (GIF CompuServe)"; }
    const char *GetExtension(Data*)
	{ return "gif"; }
    char *DoConvert(IStream &is, Object **);
    Object *Convert(Data *data, Class *want);
    bool CanConvert(Data *data, Class *want)
	{ return data->Type() == Symbol("GIF") && want == Meta(Bitmap); }
    bool CanConvertTo(Object *op)
	{ return FALSE; }
//      { return op->IsKindOf(Bitmap); }
    bool ConvertTo(Data *data, Object *op);

private:
    Point fSize;
    // decompression stuff
    unsigned char *fStackPtr;
    unsigned int *fPrefix;
    unsigned char *fSuffix;
    unsigned char *fStack;
    int fDataSize, fCodeSize, fCodeMask;
    int fClear, fEoi, fAvailable, fOldCode;

    void InitColors(Bitmap *bm, unsigned char colormap[COLORS][3], int ncolors, int background);
    bool ReadImage(DevBitmap *dbm, IStream &is, long ilmap[]);
    bool Decode(DevBitmap *dbm, int code, long &fill, long ilmap[]);
};

NewMetaImpl0(GIFConverter,Converter);

#define DECODETABSIZE 4096

//---- utils -----------------------------------------------------------

static long GetSmallEndian(IStream &is,int bytes)
{
    byte buf[4];

    if (is.read(buf, bytes) != bytes) {
	Warning("GetSmallEndian", "read(buf, bytes) != bytes");
	return 0;
    }

    switch (bytes) {
    case 1:
	return buf[0];
    case 2:
	return buf[0] + (buf[1] << 8);
    case 3:
	return buf[0] + (buf[1] << 8) + (buf[2] << 16);
    case 4:
	return buf[0] + (buf[1] << 8) + (buf[2] << 16) + (buf[3] << 24);
    default:
	Warning("IStream::GetSmallEndian", "%d out of range", bytes);
	break;
    }
    return 0;
}

//---- GIF -----------------------------------------------------------

GIFConverter::~GIFConverter()
{
    SafeDelete(fPrefix);
    SafeDelete(fSuffix);
    SafeDelete(fStack);
}

void GIFConverter::InitColors(Bitmap *bm, unsigned char colormap[COLORS][3], int ncolors, int background)
{
    RGB rgb;
    register int i;
    short gammaTable[255+1];
    double gamma= 1.7; // default 1.7; smaller makes image brighter

    for (i= 0; i <= cMaxRGB; i++)
	gammaTable[i]= int(cMaxRGB * Math::Power((double)i/(double)cMaxRGB, gamma)+0.5);

    for (i= 0; i < ncolors; i++) {
	rgb.red   = gammaTable[colormap[i][0]];
	rgb.green = gammaTable[colormap[i][1]];
	rgb.blue  = gammaTable[colormap[i][2]];
	if (i == background)
	    rgb.alpha= cMinRGB;
	else
	    rgb.alpha= cMaxRGB;
	bm->ChangeRGB(i, &rgb);
    }
}

bool GIFConverter::Decode(DevBitmap *dbm, int code, long &fill, long ilmap[])
{
    int inCode;
    static unsigned char firstChar;

    if (code == fClear) {
	fCodeSize= fDataSize + 1;
	fCodeMask= (1 << fCodeSize) - 1;
	fAvailable= fClear + 2;
	fOldCode= -1;
	return TRUE;
    }

    if (fOldCode == -1) {
	dbm->SetPixel(fill % fSize.x, ilmap[fill / fSize.x], (u_long)fSuffix[code]);
	fill++;
	firstChar= fOldCode= code;
	return TRUE;
    }
    if (code > fAvailable) {
	fprintf(stderr, "code % d to large for %d\n", code, fAvailable);
	return FALSE;
    }
    inCode= code;
    if (code == fAvailable) {
	*fStackPtr++= firstChar;
	code= fOldCode;
    }
    while (code > fClear) {
	*fStackPtr++= fSuffix[code];
	code= fPrefix[code];
    }
    *fStackPtr++= firstChar = fSuffix[code];
    fPrefix[fAvailable]= fOldCode;
    fSuffix[fAvailable]= firstChar;
    fAvailable++;
    if (((fAvailable & fCodeMask) == 0) && (fAvailable < DECODETABSIZE)) {
	fCodeSize++;
	fCodeMask+= fAvailable;
    }
    fOldCode= inCode;
    do {
	dbm->SetPixel(fill % fSize.x, ilmap[fill / fSize.x], (u_long)*(--fStackPtr));
	fill++;
    } while (fStackPtr > fStack);
    return TRUE;
}

bool GIFConverter::ReadImage(DevBitmap *dbm, IStream &is, long ilmap[])
{
    unsigned char buf[255];
    int bits= 0;
    unsigned int datum= 0;
    unsigned char *ch;
    int count, code;
    long fill= 0;

    fPrefix= new u_int[DECODETABSIZE];
    fSuffix= new u_char[DECODETABSIZE];
    fStack= new u_char[DECODETABSIZE];

    fDataSize= GetSmallEndian(is,1);
    fClear= 1 << fDataSize;
    fEoi= fClear + 1;
    fAvailable= fClear + 2;
    fOldCode= -1;
    fCodeSize= fDataSize + 1;
    fCodeMask= (1 << fCodeSize) - 1;
    for (code= 0; code < fClear; code++) {
	fPrefix[code]= 0;
	fSuffix[code]= code;
    }
    fStackPtr= fStack;
    for (count= GetSmallEndian(is,1); count > 0; count= GetSmallEndian(is,1)) {
	is.read(buf, count);
	for (ch= buf; count-- > 0; ch++) {
	    datum+= *ch << bits;
	    bits+= 8;
	    while (bits >= fCodeSize) {
		code= datum & fCodeMask;
		datum>>= fCodeSize;
		bits-= fCodeSize;
		if (code == fEoi)
		    goto exitloop;
		if (!Decode(dbm, code, fill, ilmap))
		    goto exitloop;
	    }
	}
	if (fill >= fSize.x*fSize.y) {
	    Warning("ReadImage", "too many pixels");
	    goto exitloop;
	}
    }
exitloop:
    if (fill != fSize.x*fSize.y)  {
	fprintf(stderr, "Warning: wrong rastersize: %ld bytes ", fill);
	fprintf(stderr, "instead of %ld bytes\n", (long) fSize.x*fSize.y);
    }
    return TRUE;
}

char *GIFConverter::DoConvert(IStream &is, Object **op)
{
    Bitmap *BM;
    register DevBitmap *dbm;
    int x, y, depth= 8, flags, transparent= -1;
    byte magic[6];
    bool interlaced;
    bool localCM, globalCM; // global/local color map
    long globalbits, localbits; // bits of color map entry
    unsigned char globalmap[COLORS][3]; // RGB color map
    unsigned char localmap[COLORS][3]; // RGB color map
    char ch;
    long *ilmap= 0;

    // check file
    if (is.read(magic, 6) != 6)
	return "file too short";

    if (magic[0] != 'G' || magic[1] != 'I' || magic[2] != 'F' )
	return "not a GIF file";

    if (magic[3] != '8' || magic[5] != 'a') // 87a, 89a
	return "unknown GIF version number";

    // read header
    fSize.x= GetSmallEndian(is,2);
    fSize.y= GetSmallEndian(is,2);
    if(fSize.x==0||fSize.y==0) {
	return "Corrupted GIF file";
    }
    flags= GetSmallEndian(is,1);
    GetSmallEndian(is,1);
    GetSmallEndian(is,1);
    globalCM= flags & 0x80;

    if (globalCM) {
	globalbits= (flags & 0x07) + 1;
	is.read((byte*)&globalmap, 3*(1<<globalbits));
    }
    //fprintf(stderr, "Global: w: %d h: %d f: %d\n", fSize.x, fSize.y, flags);
    ch= GetSmallEndian(is,1);
    while (ch != ';' && !is.bad()) {
	switch (ch) {
	    case '\0':
		break;
	    case ',':
		{
		x= GetSmallEndian(is,2);
		y= GetSmallEndian(is,2);
		fSize.x= GetSmallEndian(is,2);
		fSize.y= GetSmallEndian(is,2);
		flags= GetSmallEndian(is,1);
		localCM= flags & 0x80;
		if (!localCM && !globalCM)
		    return "colormap is missing";

		interlaced= flags & 0x40;
		ilmap= new long[fSize.y];
		if (interlaced) {
		    int i, j= 0;
		    for (i= 0; i < fSize.y; i+= 8)
			ilmap[j++]= i;
		    for (i= 4; i < fSize.y; i+= 8)
			ilmap[j++]= i;
		    for (i= 2; i < fSize.y; i+= 4)
			ilmap[j++]= i;
		    for (i= 1; i < fSize.y; i+= 2)
			ilmap[j++]= i;
		} else {
		    for (int i= 0; i < fSize.y; i++)
			ilmap[i]= i;
		}

		//fprintf(stderr, "Local: w: %d h: %d f: %d\n", fSize.x, fSize.y, flags);
		BM= new Bitmap(fSize, depth);
		dbm= BM->GetDevBitmap();

		bool rc= ReadImage(dbm, is, ilmap);
		delete ilmap;
		if (!rc)
		    return "data error";

		if (localCM) {
		    localbits= (flags & 0x7) + 1;
		    is.read((byte*)&localmap, 3*(1<<globalbits));
		    InitColors(BM, localmap, 1<<localbits, transparent);
		} else if (globalCM)
		    InitColors(BM, globalmap, 1<<globalbits, transparent);
		}
		break;
	    case '!':
		{
		    int count, code;
		    char buf[255];

		    code= GetSmallEndian(is,1);
		    switch (code) {
		    case 249:
			count= GetSmallEndian(is,1);
			is.read(buf, count);
			//fprintf(stderr, "x249: %d %d %d %d\n", buf[0], buf[1], buf[2], buf[3]);
			transparent= buf[3];
			break;
		    default:
			//fprintf(stderr, "GIFConverter::Extension: %d\n", code);
			while (count= GetSmallEndian(is,1))
			    is.read(buf, count);
			break;
		    }
		}
		break;
	    default:
		return "illegal GIF block type\n";
	}
	ch= GetSmallEndian(is,1);
    }
    if (BM)
	*op= BM;
    else
	*op= 0;
    return 0;
}

Object *GIFConverter::Convert(Data *data, Class*)
{
    IStream is(data->GetStreamBufForReading(), 0, TRUE);
    Object *op= 0;
    char *msg= DoConvert(is, &op);
    if (msg)
	Warning("Convert", "error in DoConvert %s", msg);
    return op;
}

bool GIFConverter::ConvertTo(Data *data, Object *op)
{
    return FALSE;
}

void RefGIFConverter()
{
    static Class *cl= Meta(GIFConverter);
    static GIFMatcher gifmatcher(cDocTypeGIF);
}
