#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma implementation
#endif

#include "Document.h"

#include "Class.h"
#include "FileDialog.h"
#include "System.h"
#include "CommandProcessor.h"
#include "WindowSystem.h"
#include "Env.h"
#include "Collection.h"
#include "CLib.h"
#include "Data.h"
#include "MenuBar.h"
#include "Menu.h"
#include "CmdNo.h"
#include "ETString.h"
#include "Application.h"
#include "Alert_e.h"
#include "TypeMatcher.h"

static FileDialog *loadDialog, *saveDialog;

//---- Document ----------------------------------------------------------------

NewAbstractMetaImpl(Document,Manager, (T(docType), TP(docData)));

Document::Document(const Symbol &dt)
{
    docType= dt;
    docData= new FileData((char*)0);
    Manager::SetName(GetBaseName());
    SetOnDismiss(eMgrIconize);
}

Document::~Document()
{
    SafeDelete(docData);
}

CommandProcessor *Document::MakeCmdProcessor()
{
    int n = Env::GetValue("Document.UndoLevel", 1);
    if (n == 1)
	return new SingleCommandProcessor;
    else
	return new UnboundedCommandProcessor;
}
    
FileDialog *Document::MakeFileDialog(FileDialogType)
{
    return gApplication->MakeFileDialog();
}

char *Document::GetBaseName()
{
    return docData->ShortName();
}

int Document::UniqueId()
{
    return docData->UniqueId();
}

bool Document::IsUntitled()
{
    return docData->IsUntitled();
}

bool Document::IsConverted()
{
//    if (docType == cDocTypeAscii && docData->IsAscii())
	return FALSE;
//    return docData->Type() != docType;
}

void Document::SetData(Data *fd)
{
    if (fd && fd != docData) {
	SafeDelete(docData);
	docData= (Data*) fd->DeepClone();
    }
    SetName(GetBaseName());
}

//---- Menus -------------------------------------------------------------------

MenuBar *Document::DoMakeMenuBar()
{
    MenuBar *mb= new MenuBar;
    mb->AddMenuRight(MakeMenu(cHELPMENU));
    mb->AddMenu(MakeMenu(cFILEMENU));
    mb->AddMenu(MakeMenu(cEDITMENU));
    mb->AddMenu(MakeMenu(cAPPMENU));
    return mb;
}

Menu *Document::MakeMenu(int menuId)
{
    Menu *m;
    
    switch (menuId) {
    case cFILEMENU:
	m= new Menu(cFILEMENU, "File");
	m->AppendItems( 
	    "New@N",        cNEW,
	    "Open ...@O",   cOPEN,
	    "-",
	    "Load ...@L",   cLOAD,
	    "Save@S",       cSAVE,
	    "Save As ...",  cSAVEAS,
	    "Revert",       cREVERT,
	    "Close",        cCLOSE,
	    "-",
	    "Print ...@P",  cPRINT,
	    "-",
	    "Quit@Q",       cQUIT,
	    0);
	break;
    
    default:
	m= Manager::MakeMenu(menuId);
    }
    return m;
}

void Document::DoSetupMenu(Menu *m)
{
    Manager::DoSetupMenu(m);
    GetCmdP()->DoSetupMenu(m);
    if (Modified()) {
	m->EnableItem(cSAVE);
	if (!IsUntitled())
	    m->EnableItem(cREVERT);
    }
    m->EnableItems(cLOAD, cSAVEAS, cPRINT, 0);
}

Command *Document::DoMenuCommand(int cmd)
{
    switch(cmd) {
    case cSAVE:     Save();     break;
    case cSAVEAS:   SaveAs();   break;
    case cLOAD:     Load();     break;
    case cREVERT:   Revert();   break;
    case cCLOSE:    Close();   break;
    default:
	return Manager::DoMenuCommand(cmd);
    }
    return gNoChanges;
}

void Document::CollectParts(Collection *col)
{
    Manager::CollectParts(col);
    if (docData)
	col->Add(docData);
}

void Document::ExtCommand(int c, char *req, char *args, int l, char *&rc, int &rl)
{
    if (strcmp(req, "load") == 0)
	Load(args);
    else if (strcmp(req, "save") == 0)
	Save();
    else
	Manager::ExtCommand(c, req, args, l, rc, rl);
}

void
Document::Finish()
{
    GetCmdP()->Finish();
}

bool Document::Modified()
{
    return GetCmdP()->Modified();
}

bool Document::SavedChanges()
{
    if (Modified()) {
	switch (ShowAlert(eAlertCaution, "Save changes to @B%s@B ?", GetBaseName())) {
	case cIdYes:
	    return Save();
	case cIdNo:
	    return TRUE;
	case cIdCancel:
	    return FALSE;
	}
    }
    return TRUE;
}

bool Document::Load(char *filename, bool unique)
{
    if (filename) {
	Data *d= new FileData(filename);
	bool b= LoadData(d, unique);
	delete d;
	return b;
    }
    return LoadData(0, unique);
}

bool Document::LoadData(Data *data, bool unique)
{
    if (! SavedChanges())
	return FALSE;

    if (data == 0) {
	if (loadDialog == 0)
	    loadDialog= MakeFileDialog(eFDTypeRead);
	if (loadDialog->ShowInWindow(eFDRead, GetWindow(), this) != cIdOk)
	    return FALSE;
	data= loadDialog->GetData();
    }

    if (CanLoad(data))
    {
	if (data->UniqueId() == 0)
        {
	    ShowAlert(eAlertNote, "Something wrong with @B%s@B\n%s",
				data->Name(), gSystem->GetErrorStr());
	    return FALSE;
	}

	if (unique)
	{
	    Document *doc= gApplication->FindDocument((int)data->UniqueId());
	    if (doc && !DoFileIsAlreadyOpen(doc, data->FullName()))
		return FALSE;
	}

	Finish();
	AboutToLoad(data);
	SetData(data);
	bool ok=DoReadData(docData);

	gWindowSystem->Update();
	return ok;
    }
    else
	gApplication->OpenDocument(data->FullName());

    return TRUE;
}

void Document::AboutToLoad(Data*)
{
}

bool Document::Close()   // return TRUE if OK
{
    if ( SavedChanges() )
	return Manager::Close();
    return FALSE;
}

bool Document::Save()
{
    if (Modified() && (IsUntitled() || IsConverted()))
	return SaveAs();

    if (! Modified()) {
	ShowAlert(eAlertNote, "No changes since last save");
	return TRUE;
    }

    if (! docData->IsWritable()) {
	ShowAlert(eAlertNote, "Document @B%s@B is not writable\n%s",
					docData->Name(), gSystem->GetErrorStr());
	return FALSE;
    }
    //trigger determination of LineSeparatorStyle (before Backup!)
    docData->GetLineSeparatorStyleNew();

    Store();
	
    return TRUE;
}

bool Document::SaveAs()
{
    if (saveDialog == 0)
	saveDialog= MakeFileDialog(eFDTypeWrite);

    if (saveDialog->ShowInWindow(eFDWrite, GetWindow(), this) == cIdOk) {
	//remember some important data from the original:
	bool isAscii; ELineSeparatorStyle lss;
	isAscii= docData->IsAscii();
	if (isAscii) lss= docData->GetLineSeparatorStyleNew();
	
	SetData(saveDialog->GetData());
	
	//reset some of the original data:
	docData->SetIsAscii(isAscii);
	if (isAscii) docData->SetLineSeparatorStyleNew(lss);
 
	Store();
	return TRUE;
    }
    return FALSE;
}

void Document::Revert()
{
    if (Modified() && ShowAlert(eAlertCaution, "Discard all changes to @B%s@B ?", 
						    GetBaseName()) == cIdYes) {
	Finish();
	Load(docData->FullName(), FALSE);
    }
}

//---- Load/Store Documents -----------------------------------------------------

void Document::MakeBackup(char *path, char *suffix)
{
    CLib::Rename(path, form("%s%s", path, suffix));
}

bool Document::CanLoad(Data *data)
{
    if (GetDocumentType() == data->Type())
	return TRUE;
    if (GetDocumentType() == cDocTypeAscii && data->IsAscii())
	return TRUE;
    return FALSE;
}

bool Document::DoFileIsAlreadyOpen(Document *shown, char *name)
{
    ShowAlert(eAlertNote, "Document @B%s@B is already open", name);
    if (shown != this)
	shown->Show();
    return FALSE;
}

void Document::Store()
{
    char *suffix = "%";
    Finish();

    if ( !docData->IsUntitled() && !gSystem->AccessPathName(docData->FullName(), 0) ) {
	MakeBackup(docData->FullName(), suffix);
    }
    OStream to(docData->GetStreamBufForWriting(), TRUE);
    if ( DoWrite(to, docData) ) {
	if ( to.good() && !Env::GetValue("Document.MakeBackup", FALSE) ) {
	     CLib::Unlink(form("%s%s", docData->FullName(), suffix));
	}
    }
}

bool Document::DoWrite(OStream& to, Data*)
{
    to << cMagic SP << docType.AsString() SP << gApplication->ProgramName() NL;
    return TRUE;
}

bool Document::DoReadData(Data *data)
{   
    IStream from(data->GetStreamBufForReading(), 0, TRUE);
    return DoRead(from, data);
}

bool Document::DoRead(IStream &from, Data*)
{
    char c;
    // overread magic cookie
    while (from.get(c))
	if (c == '\n')
	    break;
    return TRUE;
}
