#include "GridView.h"
#include "Class.h"
#include "Clipper.h"
#include "ETMath.h"

//---- CellAttribute -----------------------------------------------------------

CellAttributes::CellAttributes()
    : fInk(gInkBlack), fFont(gSysFont), fHAlign(eLeft), fVAlign(eTop)
{
}

CellAttributes::CellAttributes(Ink *i, Font *f, EAlign ha, EAlign va)
    : fInk(i), fFont(f), fHAlign(ha), fVAlign(va)
{
}

CellAttributes::CellAttributes(const CellAttributes &ca)
    : fInk(ca.fInk), fFont(ca.fFont), fHAlign(ca.fHAlign), fVAlign(ca.fVAlign)
{
}

//---- GridSelector ------------------------------------------------------------

GridSelector::GridSelector(GridView *gv, bool range, int c, int cl)
    : fGridView(gv), fCode(c), fClicks(cl)
{
    fOptions= range ? eSelRange : eSelDefault;
}

void GridSelector::TrackFeedback(Point, Point, bool)
{
}

Command *GridSelector::TrackMouse(TrackPhase atp, Point, Point, Point np)
{
    Point cell;
    if (!fGridView->PosToCell(np, &cell)) {
	if (atp == eTrackPress) {
	    fAnchor= cell;
	    fSelection= fGridView->GetSelection();
	}
	Rectangle r;
	if (fOptions & eSelRange)
	    r= NormRect(fAnchor, cell);
	else
	    r= Rectangle(cell, gPoint1);
	switch (fCode) {
	case 0:
	    fGridView->SetSelection(r);
	    break;
	case 1:
	    fGridView->SetSelection(Union(fSelection, r));
	    break;
	case 2:
	    fGridView->SetSelection(Difference(fSelection, r));
	    break;
	}
    } else
	fGridView->ClearSelection();

    if (atp == eTrackPress) {
	fAnchor= cell;
	fSelection= fGridView->GetSelection();
    }

    if (atp == eTrackRelease) {
	fGridView->DoSelect(fClicks);
	return gNoChanges;
    }
    return this;
}

//---- GridView ----------------------------------------------------------------

NewMetaImpl(GridView,View, (T(fSelection), T(fMargin), TB(fUseCache), TB(fModified)));

GridView::GridView(EvtHandler *eh, bool c)
    : fModified(TRUE), fOptions(eSelDefault), fUseCache(c), View(eh, gRect0)
{
}

GridView::GridView(EvtHandler *eh, GridSelectionOptions o, bool c)
    : fModified(TRUE), fOptions(o), fUseCache(c), View(eh, gRect0)
{
}

GridView::~GridView()
{
    SafeDelete(fGridCache);
}

Point GridView::GetColsRows()
{
    return gPoint8;
}

Coord GridView::GetGap(Direction, long)
{
    return 3;
}

Metric GridView::CellSize(const Point&)
{
    return Metric(1);
}

Coord GridView::GetColWidth(long col)
{
    register Coord w= 0;
    register long y, rows= GetRows();
    for (y= 0; y < rows; y++)
	w= Math::Max((int)w, (int)CellSize(Point((int)col, (int)y)).Width());
    return w;
}

Coord GridView::GetRowHeight(long row)
{
    register Coord h= 0;
    register long x, cols= GetCols();
    for (x= 0; x < cols; x++)
	h= Math::Max((int)h, (int)CellSize(Point((int)x, (int)row)).Height());
    return h;
}

Coord GridView::GetColOrigin(long)
{
    return 0;
}

Coord GridView::GetRowOrigin(long)
{
    return 0;
}

Coord GridView::GetRowBase(long row)
{
    register Coord b= 0;
    register int x, cols= GetCols();
    for (x= 0; x < cols; x++)
	b= Math::Max((int)b, (int)CellSize(Point((int)x, (int)row)).Base());
    return b;
}

CellAttributes *GridView::GetCellAttributes(const Point&)
{
    static CellAttributes ca;
    return &ca;
}

void GridView::DrawCell(const Point&, const Rectangle&, const Rectangle&, bool)
{
}

void GridView::DrawGridRect(Direction dir, long, const Rectangle &r)
{
    if (r.IsNotEmpty()) {
	if (dir == eHor)
	    GrLine(r.W(), r.E());
	else
	    GrLine(r.N(), r.S());
    }
}

void GridView::Draw(Rectangle r)
{
    Point p, p1, p2;
    PosToCell(r.NW(), &p1);
    PosToCell(r.SE(), &p2);
    Rectangle r1, r2, cellrect(GetGapCellRect(NormRect(p1, p2)));

    r1.origin.x= cellrect.origin.x;
    r1.extent.x= cellrect.extent.x;

    r2.origin.y= cellrect.origin.y;
    r2.extent.y= cellrect.extent.y;

    for (p.y= p1.y; p.y <= p2.y; p.y++) {
	for (p.x= p1.x; p.x <= p2.x; p.x++) {
	    Rectangle cr(GetCellOrigin(p)+GetOrigin(), GetCellExtent(p));
	    DrawCell(p, cr, Inter(cr, r), fSelection.ContainsPoint(p));
	    if (p.y == p1.y) {
		r2.extent.x= GetColGap(p.x);
		r2.origin.x= ColOrigin(p.x+1)-r2.extent.x;
		DrawGridRect(eVert, p.x, r2);
	    }
	}
	r1.extent.y= GetRowGap(p.y);
	r1.origin.y= RowOrigin(p.y+1)-r1.extent.y;
	DrawGridRect(eHor, p.y, r1);
    }
    if (!contentRect.Inset(fMargin).ContainsRect(r))
	DrawBorder(r);
}

void GridView::DrawBorder(const Rectangle &r)
{
    if (fMargin.x > 0)
	GrStrokeRect(contentRect);
}

Metric GridView::GetMinSize()
{
    return Metric(ColOrigin(GetCols())+fMargin.x-GetColGap(GetCols()-1),
		  RowOrigin(GetRows())+fMargin.y-GetRowGap(GetRows()-1),
		RowBase(0)+fMargin.y);
}

void GridView::ConstrainScroll(Point *p)
{
    //Point cell;
    //if (!PosToCell(*p, &cell))
    //        p->y= RowOrigin(cell.y);
}

bool GridView::PosToCell(const Point &pos, Point *cell)
{
    register Coord base, last, x, y, s, e;
    register bool out= FALSE;
    Point p(pos-GetOrigin()-fMargin);
    long cols= GetCols(), rows= GetRows();
    bool found;

    if (p.x < 0) {
	out= TRUE;
	x= 0;
    } else {
	found= FALSE;
	base= 0;
	last= cols-1;
	while (last >= base) {
	    x= (base + last) / 2;
	    s= ColOrigin(x);
	    e= s+ColWidth(x)+GetColGap(x);
	    if (p.x >= s && p.x < e) {
		found= TRUE;
		break;
	    }
	    if (p.x < s)
		last= x - 1;
	    else
		base= x + 1;
	}
	if (!found) {
	    out= TRUE;
	    x= cols-1;
	}
    }

    if (p.y < 0) {
	out= TRUE;
	y= 0;
    } else {
	found= FALSE;
	base= 0;
	last= rows-1;
	while (last >= base) {
	    y= (base + last) / 2;
	    s= RowOrigin(y);
	    e= s+RowHeight(y)+GetRowGap(y);
	    if (p.y >= s && p.y < e) {
		found= TRUE;
		break;
	    }
	    if (p.y < s)
		last= y - 1;
	    else
		base= y + 1;
	}

	if (!found) {
	    out= TRUE;
	    y= rows-1;
	}
    }

    if (cell) {
	cell->x= x;
	cell->y= y;
    }

    return out;
}

Rectangle GridView::GetCellRect(const Point &cell)
{
    if (cell.x < 0 || cell.x >= GetCols() || cell.y < 0 || cell.y >= GetRows())
	return gRect0;
    return Rectangle(GetCellOrigin(cell)+GetOrigin(), GetCellExtent(cell));
}

Rectangle GridView::GetCellRect(const Rectangle &cells)
{
    if (cells.IsEmpty())
	return gRect0;
    Point p= cells.origin+cells.extent;
    return NormRect(GetOrigin()+GetCellOrigin(cells.origin),
	    GetOrigin()+GetCellOrigin(p-gPoint1)+GetCellExtent(p-gPoint1));
}

Rectangle GridView::GetGapCellRect(const Point &cell)
{
    if (cell.x < 0 || cell.x >= GetCols() || cell.y < 0 || cell.y >= GetRows())
	return gRect0;
    return NormRect(GetCellOrigin(cell)+GetOrigin(),
	    GetCellOrigin(cell+gPoint1)-gPoint1+GetOrigin());
}

Rectangle GridView::GetGapCellRect(const Rectangle &cells)
{
    if (cells.IsEmpty())
	return gRect0;
    return NormRect(GetCellOrigin(cells.NW())+GetOrigin(),
	    GetCellOrigin(cells.origin+cells.extent)-gPoint1+GetOrigin());
}

void GridView::DoObserve(int, int, void*, Object*)
{
    SetModified();
    ForceRedraw();
}

bool GridView::CheckCache()
{
    if (!fUseCache)
	return FALSE;

    if (!fModified)
	return TRUE;

    fModified= FALSE;

    register Coord w;
    register long px, py;
    Point p;
    long cols= GetCols(), rows= GetRows();

    SafeDelete(fGridCache);
    fGridCache= new GridCache[Math::Max(cols, rows)+1];

    fGridCache[0].x= fMargin.x;
    fGridCache[0].y= fMargin.y;

    for (px= 0; px < cols; px++) {
	w= 0;
	for (py= 0; py < rows; py++) {
	    p.x= px;
	    p.y= py;
	    Metric m(CellSize(p));
	    w= Math::Max((int)w, (int)m.Width());
	    fGridCache[py+1].y= Math::Max((int)fGridCache[py+1].y, (int)m.Height());
	    fGridCache[py].bs= Math::Max((int)fGridCache[py].bs, (int)m.Base());
	}
	fGridCache[px+1].x= w;
    }

    for (px= 1; px <= cols; px++)
	fGridCache[px].x+= fGridCache[px-1].x + GetColGap(px-1);
    for (py= 1; py <= rows; py++)
	fGridCache[py].y+= fGridCache[py-1].y + GetRowGap(py-1);

    SetExtent(GetMinSize().extent);

    return TRUE;
}

void GridView::SetSelection(const Rectangle &newsel)
{
    SetSelection(Region(newsel));
}

void GridView::SetSelection(const Region &newrgn)
{
    Rectangle all(gPoint0, GetColsRows());
    Region newsel(Intersection(newrgn, all));
    if (!(fSelection == newsel)) {
	Rectangle r;
	Region reg(SymDifference(fSelection, newsel));
	RegionRectIter next(reg);
	while (next(r))
	    InvalidateRect(GetGapCellRect(r));
	fSelection= newsel;
	Control(GetId(), cPartSelectionChanged, &fSelection);
    }
}

void GridView::DoSelect(int clicks)
{
    if (fSelection.IsNotEmpty()) {
	int partcode= clicks >= 2 ? cPartCollDoubleSelect : cPartCollSelect;
	int y= fSelection.BBox().origin.y;
	Control(GetId(), partcode, (void*)y);
    }
}

Command *GridView::DoLeftButtonDownCommand(Point lp, Token t, int clicks)
{
    int code= 0;

    if (t.IsShift()) {
	Point cell;
	PosToCell(lp, &cell);
	if (fSelection.ContainsPoint(cell))
	    code= 2;
	else // if (fOptions & eSelExtend)
	    code= 1;
    }
    return new GridSelector(this, fOptions & eSelRange, code, clicks);
}

