#ifndef GridView_First
#define GridView_First
#ifdef __GNUG__
#pragma interface
#endif

#include "View.h"
#include "Region.h"

class Clipper;

struct GridCache {
    Coord x, y, bs;
};

enum GridSelectionOptions {
    eSelRange         = BIT(0),
    eSelStuckToBorder = BIT(1),
    eSelExtend        = BIT(2),
    eSelDefault       = 0
};

//---- CellAttribute -----------------------------------------------------------

struct CellAttributes {
    Ink *fInk;
    Font *fFont;
    EAlign fHAlign, fVAlign;

    CellAttributes();
    CellAttributes(Ink*, Font*, EAlign ha, EAlign va);
    CellAttributes(const CellAttributes &ca);
};

//---- GridView ----------------------------------------------------------------

class GridView: public View {
public:
    MetaDef(GridView);

    GridView(EvtHandler *eh, bool cache= TRUE);
    GridView(EvtHandler *eh, GridSelectionOptions options, bool cache= TRUE);
    ~GridView();

    //---- cache
    void SetModified();
    void DoObserve(int, int, void*, Object*);

    //---- querying layout
    long GetRows();
    long GetCols();
    Point GetGap(const Point &cell);
    Coord ColWidth(long col);
    Coord RowHeight(long row);
    Coord RowBase(long row);
    Coord ColOrigin(long col);
    Coord RowOrigin(long row);
    Coord GetRowGap(long row);
    Coord GetColGap(long col);

    //---- coordinates to cells
    virtual bool PosToCell(const Point &pos, Point *cell= 0);
	// returns cell containing point
    //bool RectToCell(const Rectangle &rect, Rectangle *cells= 0);
	// returns cell containing point

    //---- cells to coordinates
    Rectangle GetCellRect(const Point &cell);
    Rectangle GetCellRect(const Rectangle &cells);
    Rectangle GetGapCellRect(const Point &cell);
    Rectangle GetGapCellRect(const Rectangle &cells);
    Point GetCellOrigin(const Point &cell);
    Point GetCellExtent(const Point &cell);

    //---- selection
    virtual void DoSelect(int clicks);
    virtual void SetSelection(const Region &newsel);
    void SetSelection(const Rectangle &newsel);
    void SetSelection(const Point &cell);
    void ClearSelection(bool redraw= TRUE);
    Region GetSelection();

    //---- attributes
    virtual CellAttributes *GetCellAttributes(const Point &cell);

protected:
    //---- cache
    virtual bool CheckCache();

    //---- drawing output operations
    void Draw(Rectangle );
    virtual void DrawCell(const Point &cell, const Rectangle&, const Rectangle&, bool);
    virtual void DrawGridRect(Direction dir, long x, const Rectangle &r);
    virtual void DrawBorder(const Rectangle &r);

    //---- events
    Command *DoLeftButtonDownCommand(Point lp, Token t, int clicks);
    void ConstrainScroll(Point *p);

    //---- override these to implement layout
    virtual Point GetColsRows();
    virtual Coord GetGap(Direction dir, long colrow);
    virtual Coord GetColWidth(long col);
    virtual Coord GetRowHeight(long row);
    virtual Coord GetRowBase(long row);
    virtual Coord GetColOrigin(long col);
    virtual Coord GetRowOrigin(long row);
    virtual Metric CellSize(const Point &cell);
    Metric GetMinSize();

protected:
    GridSelectionOptions fOptions;
    Point fMargin;
private:
    Region fSelection;
    bool fUseCache, fModified;
    GridCache *fGridCache;
};

//---- GridSelector ------------------------------------------------------------

class GridSelector: public Command {
public:
    GridSelector(GridView *gv, bool r, int c, int cl);
    void TrackFeedback(Point, Point, bool);
    Command *TrackMouse(TrackPhase, Point, Point, Point);
protected:
    GridView *fGridView;
    Point fAnchor;
    int fCode, fClicks;
    Region fSelection;
    GridSelectionOptions fOptions;
};

//---- inlines -----------------------------------------------------------------

inline Region GridView::GetSelection()
    { return fSelection; }
inline void GridView::SetModified()
    { fModified= TRUE; }
inline Coord GridView::GetRowGap(long row)
    { return GetGap(eHor, row); }
inline Coord GridView::GetColGap(long col)
    { return GetGap(eVert, col); }
inline Coord GridView::ColWidth(long col)
    { return CheckCache() ? fGridCache[col+1].x-fGridCache[col].x-GetColGap(col)
			: GetColWidth(col); }
inline Coord GridView::RowHeight(long row)
    { return CheckCache() ? fGridCache[row+1].y-fGridCache[row].y-GetRowGap(row)
			: GetRowHeight(row); }
inline Coord GridView::RowBase(long row)
    { return CheckCache() ? fGridCache[row].bs : GetRowBase(row); }
inline Coord GridView::ColOrigin(long col)
    { return CheckCache() ? fGridCache[col].x : GetColOrigin(col); }
inline Coord GridView::RowOrigin(long row)
    { return CheckCache() ? fGridCache[row].y : GetRowOrigin(row); }
inline void GridView::ClearSelection(bool)
    { SetSelection(gRect0); }
inline void GridView::SetSelection(const Point &cell)
    { SetSelection(Rectangle(cell, gPoint1)); }
inline Point GridView::GetCellOrigin(const Point &cell)
    { return Point((int)ColOrigin(cell.x), (int)RowOrigin(cell.y)); }
inline Point GridView::GetCellExtent(const Point &cell)
    { return Point((int)ColWidth(cell.x), (int)RowHeight(cell.y)); }
inline Point GridView::GetGap(const Point &cell)
    { return Point((int)GetColGap(cell.x), (int)GetRowGap(cell.y)); }
inline long GridView::GetRows()
    { return (long)GetColsRows().y; }
inline long GridView::GetCols()
    { return (long)GetColsRows().x; }

#endif

