#if __GNUG__ >= 2
#  pragma implementation
#endif

#include "Class.h"
#include "HelpHandler.h"
#include "WindowSystem.h"
#include "PopupItem.h"
#include "ETString.h"
#include "Menu.h"

//---- ExternalHelpHandler ---------------------------------------------------------
#include "Application.h" /* gRpc */
#include "StreamConnection.h"
#include "CLib.h"
#include "OrdColl.h"

//---- HelpMessageString ---------------------------------------------------------
// Very simple growing string.
class HelpMessageString {
public:
    HelpMessageString();
    ~HelpMessageString();
    char* AsString()
	{ fString[fLen]=0; return fString; }
    void Reset()
	{ fLen=0;}
    void Append(const char*);
    void Append(char c);
protected:
    void Expand();
    char* fString;
    int fLen;
    int fAllocSize;
};

HelpMessageString::HelpMessageString()
{
    fAllocSize=64;
    fString=CLib::CAlloc(fAllocSize,sizeof(char));
}

HelpMessageString::~HelpMessageString()
{
    CLib::Free(fString);
}

void HelpMessageString::Expand()
{
    fAllocSize*=2;
    fString=CLib::ReAlloc(fString,fAllocSize*sizeof(char));
}

void HelpMessageString::Append(const char* str)
{
    int i;
    char prev=0;
    for(i=0; str[i]; i++) {
	char c=str[i];
	// replace special characters with '_'
	if ( !Isalnum(c) && strchr(":+-_",c)==0 ) 
	    c='_';
	if (c!='_' || prev!=c ) // no duplicate '_'
	    Append(c);
	prev=c;
    }
}

void HelpMessageString::Append(char c)
{
    if(fLen+1>=fAllocSize)
	Expand();
    fString[fLen]=c;
    fLen++;
    // string is terminated with 0 in AsString :-)
}

//---- ExternalHelpHandler ---------------------------------------------------------
class ExternalHelpHandler: public HelpHandler {
public:
    MetaDef(ExternalHelpHandler);
    ExternalHelpHandler();
protected:
    void ShowHelp(const char*);
    bool IsHelpEnabled();
    const char* MakeHelpString(Object* op);
    void BuildHelpMessage(SeqCollection *coll);
protected:
    char* fHelpTool;
    HelpMessageString fMessage;
};

//---- ExternalHelpHandler ---------------------------------------------------------
NewMetaImpl(ExternalHelpHandler,HelpHandler,(TS(fHelpTool)));

ExternalHelpHandler::ExternalHelpHandler()
{
    fHelpTool=strsave(CLib::Getenv("ET_HELP"));
}

bool ExternalHelpHandler::IsHelpEnabled()
{
    return fHelpTool && *fHelpTool;
}

void ExternalHelpHandler::ShowHelp(const char*str)
{
    gRpc->Talk(fHelpTool, "show %s", str);
}

void ExternalHelpHandler::BuildHelpMessage(SeqCollection *coll)
{
    fMessage.Reset();
    fMessage.Append("help:");
    EvtHandler *h;
    Iter next(coll,cIterBackward);
    while(h=Guard(next(),EvtHandler)) {
	fMessage.Append('/');
	fMessage.Append(form("%s",h->ClassName()));
	if(h->GetId()>0)
	    fMessage.Append(form("+%d",h->GetId()));
	char buf[32];
	buf[0]=0;
	h->InspectorId(buf,sizeof(buf)-1);
	if(buf[0])
	    fMessage.Append(form("+%s",buf));
    }
}

const char* ExternalHelpHandler::MakeHelpString(Object* op)
{
    SeqCollection* coll=new OrdCollection;
    EvtHandler *h=0;
    if(op->IsKindOf(EvtHandler))
	h=(EvtHandler*)op;
    // iterate over the event handler chain to get the object path
    while ( h ) {
	// put into list if leaf or not in exclusion list...
	if(coll->Size()==0
	   || !strismember(h->ClassName(),
			"Box","HBox","VBox",
			"Clipper","Scroller","Form","Window",
			//"Cluster","ManyOfCluster","OneOfCluster",
			"Matte","MenuBar",
			"Expander","HExpander","VExpander",
			0)
	    )
	{
	    coll->Add(h);
	}
	h=h->GetNextHandler();
    }
    if (!coll->FindPtr(gApplication))
	coll->Add(gApplication);
    BuildHelpMessage(coll);
    SafeDelete(coll);
    return fMessage.AsString();
}

//---- HelpHandler ---------------------------------------------------------
NewMetaImpl(HelpHandler,RefCounted,(T(fInContextHelp)));

HelpHandler* gHelp=0;

HelpHandler::HelpHandler()
{
    fInContextHelp=0;
}

void HelpHandler::SetHandler(HelpHandler*help)
{
    AssignRef(&gHelp,help);
}

void HelpHandler::Init()
{
    SetHandler(new ExternalHelpHandler());
}

bool HelpHandler::Help(Object* op)
{
    if(!IsHelpEnabled())
	return FALSE;
    if(fInContextHelp && op->IsKindOf(EvtHandler)) {
	EvtHandler *e=(EvtHandler*)op;
	if(e->FindNextHandlerOfClass(Meta(MenuButton))) {
	    if(!e->FindNextHandlerOfClass(Meta(Menu))) 
		return FALSE;
	}
    }
    DoHelp(op);
    EndContextHelp();
    return TRUE;
}

void HelpHandler::StartContextHelp()
{
    if(!IsHelpEnabled())
	return;
    if(fInContextHelp==0) {
	gWindowSystem->SetWait(eCrsHelp);
	fInContextHelp++;
    }
}

bool HelpHandler::InContextHelp()
{
    return bool(fInContextHelp);
}

void HelpHandler::EndContextHelp()
{
    if(!IsHelpEnabled())
	return;
    if(fInContextHelp>0) {
	fInContextHelp--;
	gWindowSystem->ResetWait();
    }
}

bool HelpHandler::HelpOnMenu(EvtHandler*op,int cmd)
{
    if(!IsHelpEnabled())
	return FALSE;
    if(fInContextHelp) {
	DoHelp(op);
	EndContextHelp();
	return TRUE;
    }
    return FALSE;
}

bool HelpHandler::IsHelpEnabled()
{
    return FALSE;
}

void HelpHandler::ShowHelp(const char*str)
{
    Error("Help",(char*)str);
}

const char* HelpHandler::MakeHelpString(Object* op)
{
    char buf[32];
    buf[0]=0;
    op->InspectorId(buf,sizeof(buf)-1);
    return form("No Help for @B%s@B (@I%s@I)",op->ClassName(),buf);
}

void HelpHandler::DoHelp(Object*op)
{
    ShowHelp(MakeHelpString(op));
}
