#ifndef TextViewProxy_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define TextViewProxy_First

#include "RefCounted.h"

//----------------------------------------------------------------------------
// THE PROBLEM
// ===========

// Let's assume a text is displayed in two TextViews, tv1 and tv2. A user edits
// the text in tv1, then loads another text into tv1. Now he invokes UNDO in
// tv2. OOPS the new text in tv1 gets modified.

// The reason is, TextCommand keeps a reference to the TextView. The undo is
// then applied using tv1.


// PROPOSAL 1
// ==========

// Use a proxy for the TextView in the TextCommand. The default implementation
// of the TextViewProxy would just keep a pointer to the view. To make this
// change as smooth as possible, TextCommand would use a SmartTextView which
// would require only minimal changes int the constructor of TextCommand.

// But how does this solve the problem? 

// (look at the code below to understand the answers)

// - The application could just keep track of the text and the
// TextViews. Whenever a text is detached from the TextView the Application-,
// Document- or Manager-specific TextViewProxy could find a TextView, that
// currently holds the text.

// - If the text is the 'stable' thing, a TextViewProxy could keep a reference
// to the text and scan all the observers of the text for a TextView....

// - Let's assume the text is displayed in multiple Managers (sharing a
// CommandProcessor), each Manager having its own TextView. The Manager that
// performs the command could tell its TextViewProxys (directly or indirectly)
// which is the current TextView.

// I think, the advantage of such a TextViewProxy is, that it decouples the
// TextCommand from the Text *and* the TextView.

// PROPOSAL 2
// ==========

// Instead of using one global TextViewProxyFactory each TextView would
// (potentially) have its own TextViewProxyFactory:

// class TextView {
// 	...
// 	TextViewProxyFactory *_tvf;
// protected:
// 	virtual TextViewProxyFactory *MakeTextViewProxyFactory()
// 		{ return gTextViewProxyFactory;} // it's a refcounted :-)
// 						 // solitair

// 	TextViewProxyFactory *GetTextViewProxyFactory()
// 		{ if(!_tvf)
// 		  	SetTextViewProxyFactory(MakeTextViewProxyFactory());
// 		  return _tvf;
// 		}
// public:
// 	~TextView()
// 		{ ... SafeUnref(_tvf); ...}
// 	TextViewProxy* MakeProxy()
// 		{ GetTextViewProxyFactory()->MakeProxy(this);}

// 	void SetTextViewProxyFactory(TextViewProxyFactory*tvf)
// 		{ AssignRef(&_tvf,tvf);}
// };


// and

// SmartTextView::SetTextView(TextView*tv)
// {
// 	SetProxy(tv->MakeProxy());
// }

// One advantage of this solution is, that the programmer can choose for each
// TextView which TextViewProxyFactory to use. For preexisting classes he would
// just use SetTextViewProxyFactory to change the behaviour, for his own
// subclasses he could overwrite MakeProxy.

// Michael
//----------------------------------------------------------------------------

class TextView;

//------------------- TextViewProxy ------------------------------------------
class TextViewProxy: public RefCounted { 
public:
	TextViewProxy(){}
	MetaDef(TextViewProxy);

    // abstract!
	virtual TextView* GetTextView();
};

//------------------- DefaultTextViewProxy ------------------------------------
// keeps a reference to a TextView

class DefaultTextViewProxy: public TextViewProxy {
	TextView* _tv;
public:
	MetaDef(DefaultTextViewProxy);
	DefaultTextViewProxy(TextView*tv)
		{_tv=tv;}
	TextView* GetTextView()
		{return _tv;}
};

//-------------------- SmartTextView ------------------------------------------
// this class is used instead of TextView* for TextCommands...

class SmartTextView {
public:
    // this is only public to be used in MetaImpl!!!!!
	// Don't touch it, it is supposed to be private.
	TextViewProxy*_p;
public:
	SmartTextView(TextView *tv=0)
		{_p=0;SetTextView(tv);}
	~SmartTextView()
		{SafeUnref(_p);}
	void SetProxy(TextViewProxy *p)
		{AssignRef(&_p,p);}
	void SetTextView(TextView *tv);

	// some nice operator overloading :-)
	TextView* operator->()
		{return _p->GetTextView();}
	operator TextView*()
		{return _p->GetTextView();}
};

//-------------------- TextViewProxyFactory -----------------------------------

class TextViewProxyFactory: public RefCounted  {
public:
	MetaDef(TextViewProxyFactory);
	TextViewProxyFactory(){}

	// default: return new DefaultTextViewProxy :-)
	virtual TextViewProxy* MakeProxy(TextView*tv);
};

#endif     
