#ifdef __GNUG__
#pragma implementation
#endif

#include "UnixSystem.h"

#include "Class.h"
#include "ETString.h"
#include "CLib.h"
#include "Error.h"
#include "UnixPtty.h"
#include "ClassManager.h"
#include "Env.h"
#include "Exception.h"
#include "ETMath.h"
#include "system.h"
 
#if !defined(OS_LINUX) && !defined(OS_SUNOS) && !defined(OS_HPUX) && !defined(OS_ULTRIX) && !defined(OS_SCO)
#include <sys/select.h>
#endif

const char *path_separator = "/";
const char *path_parentdir = "../";
char *gSessionId=0;

extern "C" {
    char *getenv(const char *);
#if !defined(OS_SUNOS)
    int mkdir(char *path, int mode);
#else
    int mkdir(const char *path, unsigned short mode);
#endif
    void _exit(int code);
    int system (const char *cmd);
    int atoi(const char *);
}
 
//---- main --------------------------------------------------------------------
#undef main
int main(int argc, char *argv[])
{
    //et3off:ETInit(argc, argv);//but that is also done in Application::Application()
				//and most sniff-apps do it themselves!
    int rc= ETMain(argc, argv);
    //et3off:ETCleanup();	//nobody seems to care!
    return rc;
}

//---- signalHandler -------------------------------------------------------

static bool asignal, signals[MAXSIGNALS];

static void sigHandler(Signals sig)
{
    switch (sig) {
    case eSigBus:
    case eSigSegmentationViolation:
    case eSigIllegalInstruction:
	throw(sig);
	break;
    default:
	break;
    }

    asignal= TRUE;
    signals[sig]= asignal= TRUE;
    if (gSystem)
	((UnixSystem*)gSystem)->DispatchSignals(TRUE);
}

//---- ZombieHandler -------------------------------------------------------

class ZombieHandler : public SysEvtHandler {
    Collection *zombieHandler;    
public:
    MetaDef(ZombieHandler);
    ZombieHandler(Collection *zh) : SysEvtHandler(eSigChild)
	{ zombieHandler= zh; }
    void Notify(SysEventCodes, int);
};

NewMetaImpl(ZombieHandler,SysEvtHandler, (TP(zombieHandler)));

void ZombieHandler::Notify(SysEventCodes, int)
{
    int pid;
    while ((pid= unix_waitchild()) > 0) {
	Iter next(zombieHandler);
	SysEvtHandler *re;
	while (re= (SysEvtHandler*) next()) {
	    if (re->GetResourceId() == pid) {
		re->Notify(eSysEvtZombie, pid);
		re->Remove();
	    }
	}
    }
}


//---- UnixSystem -------------------------------------------------------
static char prog_name[256];
void InitSystem(char *fullname)
{
    if (gSystem)
	return;
	
    if (! gInMain) {
	fprintf(stderr, "--- InitSystem called before ETInit\n"); 
	CLib::Abort();
    }
    
    gSystem= new UnixSystem();
    if (gSystem->Init()) {
	fprintf(stderr, "can't init operating system\n");
	_exit(1);
    }
    if (fullname) {
	strcpy(prog_name, gSystem->BaseName(fullname));
	gProgname = prog_name;
    }

    gProgPath= strsave(gSystem->Which(CLib::Getenv("PATH"), gProgname));
}

UnixSystem::UnixSystem() : System ("UNIX")
{
  work_dir = new char [1024];
}

UnixSystem::~UnixSystem()
{
    unix_resetsignal();
  delete work_dir;
}

void UnixSystem::AddHandler(SeqCollection *s, SysEvtHandler *h)
{
    System::AddHandler(s, h);
    if (s == asyncSignalHandler || s == signalHandler)
	unix_signal((Signals)h->GetResourceId(), sigHandler);
}

bool UnixSystem::Init()
{
    if (System::Init())
	return TRUE;

    unix_initialize();
    strcpy (work_dir, unix_initialworkingdirectory());

    AddSignalHandler (new ZombieHandler (zombieHandler));

    return FALSE;
}

void UnixSystem::SetUpMask(fd_set *m, Collection *hl, SysEventCodes c)
{
    Iter next(hl);
    register SysEvtHandler *re;

    while (re= (SysEvtHandler*) next()) 
	if (re->HasInterest(c)) {
	    FD_SET(re->GetResourceId(), m);
	    maxfd= Math::Max(maxfd, re->GetResourceId());
	}
}

void UnixSystem::DispatchReadWrite(fd_set *m, Collection *hl, SysEventCodes c)
{
    Iter next(hl);
    register SysEvtHandler *re;

    while (re= (SysEvtHandler*) next()) {
	int fd= re->GetResourceId();
	if (FD_ISSET(fd, m))
	    re->Notify(c, fd);
    }
}

void UnixSystem::DispatchSignals(bool async)
{
    Collection *shl= async ? asyncSignalHandler : signalHandler;
    Iter next(shl);
    register SysEvtHandler *re;
    int sig;

    while (re= (SysEvtHandler *)next()) {
	sig= re->GetResourceId();
	if (signals[sig] && re->HasInterest(eSysEvtSignal)) {
	    signals[sig]= FALSE;
	    re->Notify(eSysEvtSignal, sig);
	}
    }
}

void UnixSystem::InnerLoop(int timeout)
{
    fd_set readready, writeready;
    int nfds;

    if (workHandler && workHandler->Size() > 0) {
	workHandler->ForEach(SysEvtHandler,Notify)(eSysEvtWork, 0);
    }
    if(Work())
	timeout=0;

    if (asignal && signalHandler->Size() > 0) {
	DispatchSignals(FALSE);          // poll interrupts
	asignal= FALSE;
    }

    maxfd= 0;
    FD_ZERO(&readready);
    FD_ZERO(&writeready);
    if (fileInputHandler->Size() > 0)
	SetUpMask(&readready, fileInputHandler, eSysEvtRead);
    if (fileOutputHandler->Size() > 0)
	SetUpMask(&writeready, fileOutputHandler, eSysEvtWrite);

    nfds= unix_select(maxfd+1, &readready, &writeready, timeout);

    switch (nfds) {
    case -2:    // interrupt
	break;

    case -1:    // error
	SysError("UnixSystem::InnerLoop", "unix_select");
	break;

    case 0:     // timeout
	if (timeout && timeoutHandler->Size() > 0) {
	    Iter next(timeoutHandler);
	    register SysEvtHandler *eh;
	    while (eh= (SysEvtHandler*) next())
		if (eh->HasInterest(eSysEvtTimeout))
		    eh->Notify(eSysEvtTimeout, 0);
	}
	break;

    default:
	if (fileInputHandler->Size() > 0)
	    DispatchReadWrite(&readready, fileInputHandler, eSysEvtRead);
	if (fileOutputHandler->Size() > 0)
	    DispatchReadWrite(&writeready, fileOutputHandler, eSysEvtWrite);
	break;
    }
}

bool UnixSystem::CanRead(int fd, int timeout)
{
    fd_set readready;
    int rval;
		   
    if (timeout < 0) // block
	return FALSE;
    for (;;) {
	FD_ZERO(&readready);
	FD_SET(fd, &readready);
	if ((rval= unix_select(fd+1, &readready, 0, timeout)) < 0) {
	    if (rval == -1)
		SysError("UnixSystem::CanRead", "unix_select");
	    continue;
	}
	if (rval == 0) // timeout
	    return TRUE;
	return FALSE;
    }
}

bool UnixSystem::CanWrite(int fd, int timeout)
{
    fd_set writeready;
    int rval;
		   
    if (timeout < 0) // block
	return FALSE;
    for (;;) {
	FD_ZERO(&writeready);
	FD_SET(fd, &writeready);
	if ((rval= unix_select(fd+1, 0, &writeready, timeout)) < 0) {
	    if (rval == -1)
		SysError("UnixSystem::CanWrite", "unix_select");
	    continue;
	}
	if (rval == 0) // timeout
	    return TRUE;
	return FALSE;
    }
}


bool UnixSystem::AccessPathName(char *path, int mode)
{
    if (unix_fileaccess(path, mode) == 0)
	return FALSE;
    errorstr= sys_errlist[errno];
    return TRUE;
}

extern bool statCacheOn;
 
bool UnixSystem::CreateDirectoryWithParents(char *path, int mode)
{
    char *cmd = form("mkdir -p -m %o %s", mode, path);
    int ret = RunProgSynch(cmd);
    if (statCacheOn)
    	{
    	// switch off/on to clean hash database
    	StatCache(FALSE);
    	StatCache(TRUE);
    	}
    if (ret == 0)
	return FALSE;
    errorstr= sys_errlist[errno];
    return TRUE;
}

bool UnixSystem::CreateDirectory(char *path, int mode)
{
    int ret = mkdir(path, mode);
    if (statCacheOn)
    	{
    	// switch off/on to clean hash database
    	StatCache(FALSE);
    	StatCache(TRUE);
    	}
    if (ret == 0)
	return FALSE;
    errorstr= sys_errlist[errno];
    return TRUE;
}

bool UnixSystem::ChangeDirectory(char *path)
{
    bool ret;
    char rel_path[1024];
    char new_path[1024];
    char *begin;
    char *end;

    while (*path && Isspace(*path))
        path++;

    strcpy(rel_path, path);

    if (rel_path[0] == '/')
	strcpy(new_path, "/");
    else
	strcpy(new_path, work_dir);

    /* new_path: /foo/bar */
    /* rel_path: ..//xxx/yyy */

    for (begin = rel_path; *begin; )
    {
	/*  ..//xxx/yyy     ..\0\0xxx/yyy
	 *  ^                     ^
	 *  begin                 begin
	 */

	for (end = begin; *end && *end != '/'; )
	    end++;

	/*  ..//xxx/yyy     ..\0\0xxx/yyy
	 *    ^                      ^
	 *    end                    end
	 */

	while (*end == '/')
	    *end++ = 0;

	/*  ..\0\0xxx/yyy   ../\0\0xxx\0yyy
	 *        ^                     ^
	 *        end                   end
	 */

	if (begin[0] && strcmp(begin, ".") != 0)
	{
	    if (strcmp(begin, "..") == 0)
	    {
		char *ptr;
		ptr = strrchr(new_path, '/');
		if (ptr == new_path)
		    ptr++;
		*ptr = 0;
	    }
	    else
	    {
		if (strcmp(new_path, "/") != 0)
		    strcat(new_path, "/");
		strcat(new_path, begin);
	    }
	}

	begin = end;
    }

    if (strcmp(new_path, work_dir) == 0)
        return TRUE;

    ret = (bool) (chdir(new_path) == 0);
    if (ret)
        strcpy(work_dir, new_path);
    else
        errorstr = sys_errlist[errno];

    return ret;
}

 
bool 
UnixSystem::RemoveEmptyDirectory(char *path)		// removes an empty directory
{
    Dir dir(path);
    char *nm = dir();
    if (nm == 0)	// no entries -> it is not a directory !!!
		return FALSE;
    return (rmdir(path) == 0);
}

bool
UnixSystem::RemoveAllFilesInDirectory(char *path)
{
    Dir dir(path);
	int ret = 0;
    char *nm;
    int i;
    for (i= 0; nm= dir(); i++)
	if (strcmp(nm, ".") && strcmp(nm, ".."))
	    ret += unlink(nm);
    
    if (i == 0 || ret)	// an error occured during unlink
	return FALSE;
 
    return TRUE;
}

char *UnixSystem::WorkingDirectory()
{
    return work_dir;
}

char *UnixSystem::HomeDirectory()
{
    return unix_homedirectory(0);
}

char *
UnixSystem::BaseName(char *name) 
{
    if (name) {
	if (name[0] == '/' && name[1] == '\0')
	    return name;
	char *cp;
	if (cp= strrchr(name, '/'))
	    return ++cp;
	return name;
    }
    return "";
}
 
char *UnixSystem::DirName(char *pathname) 
{
    if (!pathname)
	return 0;
    if (strchr(pathname, '/'))
    {
	static char buf[1000];
	strcpy(buf, pathname);
	char *r= strrchr(buf, '/');
	if (r != buf)
	    *r= '\0';
	if (r == buf)
	{
	    r++;
	    *r = '\0';
	}
	return buf;
    }
    return ".";		//ruedi: what's that good for!?!
}
 
// a circular formating buffer:
#define PATH_CB_SIZE (1024*8)
#define PATH_FLD_SIZE 1024
static char     pathbuf[PATH_CB_SIZE];       // some slob for form overflow
static char*    pathbfree= pathbuf;
static char*    pathendbuf= &pathbuf[PATH_CB_SIZE-1];
//char *UnixSystem::ConcatPathNames(const char *va_(first), ...)
//{
//    char *buf= pathbfree;
// 
//    va_list ap;
//    va_start(ap, va_(first));
// 
//    if (buf + PATH_FLD_SIZE > pathendbuf)  
//	buf = pathbuf;
//    char *str = (char*)va_(first);
//    strcpy(buf, str);
//    while (str = va_arg(ap, char*))
//    {
//	if (buf[strlen(buf)-1] != '/')
//	    strcat(buf,"/");
//	strcat(buf,str);
//    }
//    int ll = strlen(buf);
//    if (buf[ll-1] == '/')
//	buf[ll-1] = '\0';
//    if (ll > PATH_FLD_SIZE)  // oops
//	Fatal("ConcatPathNames", "buffer overflow");
//    pathbfree= buf + ll + 1;
//    va_end(ap);
//    return StrTrim(buf);
//}
char *UnixSystem::ConcatPathNames(const char *first, ...)
{
    va_list ap;
    char *buf;
    char *str;
	int	len;
 
    va_start(ap, first);
    buf= pathbfree;
 
    if (buf + PATH_FLD_SIZE > pathendbuf)  
	buf= pathbuf;
    str= (char*)first;
    strcpy(buf, str);
    while(str = va_arg(ap, char*)) {
		len = strlen(buf);
		if (len > 0 && buf[len-1] != '/')
	    strcat(buf,"/");
	strcat(buf,str);
    }
    len = strlen(buf);
    if (len > 0 && buf[len-1]=='/') {
	if (BaseName(buf)!=buf) buf[len-1] = '\0'; //not at the root!
    }
    if (len > PATH_FLD_SIZE)  // oops
		Fatal("ConcatPathNames", "buffer overflow");
    pathbfree= buf + len + 1;
    va_end(ap);
    return StrTrim(buf);
}
  
const char *
UnixSystem::GetServerPath()
{
    return "/tmp/.sniffdir2";
}

bool UnixSystem::IsAbsolutePath(char *dir)
{
    if (dir && *dir == '/')
	return TRUE;
    return FALSE;
}
 
Directory *UnixSystem::MakeDirectory(char *name)
{
    return new UnixDirectory(name);
}
    
int UnixSystem::GetPathInfo(char *path, u_long *id, u_long *size, u_long *flags,
								u_long *modtime)
{
    return unix_filestat(path, id,  size, flags, modtime);
}
 
void UnixSystem::StatCache(bool sw)
{
    statCache(sw);
}

//---- running programs -------------------------------------
u_int UnixSystem::Sleep(u_int seconds)
{
    return sleep(seconds);
}
 
int UnixSystem::RunProgSynch(const char *cmdline)
{
    return system(cmdline);
}
 
void UnixSystem::RunProgAsynch(const char *cmdline)
{
    system(form("%s &", cmdline));
}
 
PttyConnection *UnixSystem::MakePttyConnection(char *prog, char **args)
{
    return new UnixPttyConnection(prog, args);
}

char *UnixSystem::Getenv(char *name)
{
    return CLib::Getenv(name);
}

void UnixSystem::Setenv(char *name, char *value)
{
    CLib::Putenv(strprintf("%s=%s", name, value));
}

char *UnixSystem::GetError()
{
    if (errno < 0 || errno >= sys_nerr)
	return form("errno out of range %d", errno);
    return sys_errlist[errno];
}

//---- expand the metacharacters as in the shell ------------------------

// only ~ and ~user as well as $var are expanded
// no filename globbing with wildards like *, ?, [], etc.

bool UnixSystem::ExpandPathName(char *buf, int buflen)
{
    static char error_msg[50];
    char path[1024];
    int p;		// index into "path"
    int b;		// index into "buf"

    strcpy(path, buf);
    p = b = 0;

    // skip leading white space
    while ( path[p] == ' ' || path[p] == '\t' )
	p++;

    // replace ~ or ~user with home directory
    if ( path[p] == '~' ) {
	// extract user name
	char user_name[64];
	p++;
	int u = 0;
	while ( path[p] != '\0' && path[p] != '/' && path[p] != '$' )
	    user_name[u++] = path[p++];
	user_name[u] = '\0';
	// get home directory
	char *home_dir = unix_homedirectory(user_name[0] ? user_name : 0);
	if ( home_dir == 0 ) {
	    sprintf(error_msg, "unknown user %s", user_name);
	    errorstr = error_msg;
	    return TRUE;
	} else {
	    strcpy(&buf[b], home_dir);
	    b += strlen(home_dir);
	}           
    }

    while ( path[p] != '\0' ) {
	if ( path[p] == '$' ) {
	    char envvar[256];
	    char *val;
	    int saved_p = p;
	    int e = 0;
	    p++;
	    while ( path[p] != '\0' && path[p] != '/' && path[p] != '$' )
		envvar[e++] = path[p++];
	    envvar[e] = '\0';
	    if ( envvar[0] && (val = getenv(envvar)) && val[0] ) {
		strcpy(&buf[b], val);
		b += strlen(val);
	    } else {
		p = saved_p;
		buf[b++] = path[p++];
	    }
	} else
	    buf[b++] = path[p++];
    }
    buf[b] = '\0';

    return FALSE;
}

void UnixSystem::Wait(u_int duration)
{
    unix_select(0, 0, 0, duration);
}

int UnixSystem::GetPid()
{
    return ::getpid();
}

char *UnixSystem::Which(char *search, char *file, int mode)
{
    static char name[1024];
    register char *ptr, *next;

    if (file[0] == '/') {
	if (unix_fileaccess(file, mode) == 0)
	    return file;
	return 0;
    }

    if (search == 0)
	search= ".";

    for (ptr= search; *ptr;) {
	for (next= name; *ptr && *ptr != ':'; )
	    *next++= *ptr++;
	*next= '\0';
	   
	if (name[0] != '/') {
	    char tmp[1024];
	    strcpy(tmp, name);
	    strcpy(name, work_dir);
	    strcat(name, "/");
	    strcat(name, tmp);
	}
	strcat(name, "/");
	strcat(name, file);
	   
	if (unix_fileaccess(name, mode) == 0)
	    return name;
	if (*ptr)
	    ptr++;
    }
    return 0;
}


//---- RPC --------------------------------------------------------------

int UnixSystem::OpenConnection(char *server, char *service)
{
    int fd;
    bool started= FALSE;
    
    char *appcomm= Env::GetValue("System.AppComm", "appcomm");
    if (appcomm == 0)
	return -1;
	
    char *sessionName;
    if(strchr(server,'_')==NULL) // not yet transformed
	sessionName=form("%s_%s",server,gSessionId); // needed for sniff and sniffgdb
    else sessionName=server;
    
    for (int i= 0; i < 6; i++) {
	fd= unix_UnixConnect(sessionName);
	if (fd >= 0)
	    return fd;
	if (!started) {
	    gSystem->RunProgAsynch(form("%s -s %s", appcomm, gSessionId));
	    started= TRUE;
	}
	gSystem->Sleep(i < 3 ? 1 : i*2);
    }
    return -1;
}

int UnixSystem::AnnounceTcpService(char *servicename)
{
    return unix_TcpService(servicename);
}

int UnixSystem::AnnounceUnixService(char *servername)
{
    return unix_UnixService(servername);
}

int UnixSystem::AcceptConnection(int sock)
{
    return unix_accept (sock);
}

int UnixSystem::ConnectUnixService(char *service)
{
    return unix_UnixConnect(service);
}

int UnixSystem::ConnectTcpService(char *hostname, char *service)
{
    return unix_TcpConnect(hostname, service);
}

char *UnixSystem::GetSessionName(int argc, char **argv, char *server)
{
    InitSession(argc,argv);
    return form("%s_%s",server,gSessionId);
}
 
void UnixSystem::InitSession(int argc, char **argv)
{
    for (int i = 1; i < argc; i++) {
	if (argv[i][0] == '-' && argv[i][1] == 's' && (i+1) < argc && strlen(argv[i+1]) > 0){
	    i++;
	    gSessionId=strsave(argv[i]);
	    i++;
	}
    }
    if(gSessionId == 0) { // loook for environment variable
	char *envStr=CLib::Getenv("SNiFF_SESSION_ID");
	if(envStr)
	    gSessionId=strsave(envStr);
    }
    if(gSessionId == 0) { // use default
	gSessionId=strsave("session0");
    }
 
}

//---- UnixDirectory -------------------------------------------------

UnixDirectory::UnixDirectory(char *name) : Directory(name)
{
    dirp= opendir(name);
}

UnixDirectory::~UnixDirectory()
{
    if (dirp) {
	closedir(dirp);
	dirp= 0;
    }
}

char *UnixDirectory::operator()()
{
    if (dirp)
	return unix_getdirentry(dirp);
    return 0;
}
