#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma implementation
#endif

#include "XColorMap.h"
#include "XWinSystem.h"

//---- XColorIndexMapper ------------------------------------------------

XColorIndexMapper::XColorIndexMapper(XColormap cm, int colormap_size)
    : WindowColorMap(/* writable = */ (bool)FALSE)
{
    cmap = cm;
    ncolors = colormap_size;
    local_colormap = FALSE;
    color_table = new color_cell_t[ncolors];
    // initialize "color_table" with black
    u_long pixel = XWinSystem::black;
    color_table[pixel].server = color_table[pixel].client = RGB(0,0,0);
    color_table[pixel].is_valid = TRUE;
    // and with white
    pixel = XWinSystem::white;
    color_table[pixel].server = color_table[pixel].client = RGB(255,255,255);
    color_table[pixel].is_valid = TRUE;
}

XColorIndexMapper::~XColorIndexMapper()
{
    delete color_table;
}
void XColorIndexMapper::DevFreeRGB(RGB *rgb)
{
unsigned long pixels[1];

if (!local_colormap)
    {
    // First look for an exact match in already allocated color cells.
    for ( int index = 0; index < ncolors; index++ ) {
	if ( color_table[index].is_valid
	    && ( *rgb == color_table[index].server
		|| *rgb == color_table[index].client) ) {
	    break;
	}
    }
    if((index < ncolors) && (color_table[index].ref_cnt > 0))
      color_table[index].ref_cnt--;

    // Free it only when in color_table
    // Don't free the colors from 0-7 
    // may be used by xinitcolormap
    if((index < ncolors) &&
	(index  > 7) &&
	    (index != XWinSystem::black) &&
		(index != XWinSystem::white) &&
		      (color_table[index].ref_cnt == 0))
	{
	pixels[0] = index;
	if(XFreeColors(XWinSystem::display, cmap, pixels, 1, 0))
	    color_table[index].is_valid = FALSE;
	}
    }
}

u_long
XColorIndexMapper::DevRGB2Index(RGB *rgb, RGB *result)
{
    bool found = FALSE;
    int pixel;
    int i;

    // First look for an exact match in already allocated color cells.
    for ( pixel = 0; pixel < ncolors; pixel++ ) {
	if ( color_table[pixel].is_valid
	    && ( *rgb == color_table[pixel].server
		|| *rgb == color_table[pixel].client) ) {
	    found = TRUE;
	    break;
	}
    }

    if ( ! found && ! local_colormap ) {
	// No exact match has been found.
	// Try to allocate a new shareable color cell.
	XColor xcolor;
	xcolor.flags = (DoRed | DoGreen | DoBlue);
	xcolor.red   = (rgb->red << 8) | rgb->red;
	xcolor.green = (rgb->green << 8) | rgb->green;
	xcolor.blue  = (rgb->blue << 8) | rgb->blue;

	if ( XAllocColor(XWinSystem::display, cmap, &xcolor) ) {
	    // Got a new shareable color cell.
	    // Record it in our table.
	    pixel = xcolor.pixel;
	    color_table[pixel].server = RGB(xcolor.red >> 8, xcolor.green >> 8,
					    xcolor.blue >> 8);
	    color_table[pixel].client = *rgb;
	    color_table[pixel].is_valid = TRUE;
	    found = TRUE;
	} else {	// fetch the contents of the X colormap
	    u_long i;
	    XColor *xc = new XColor[ncolors];
	    for ( i = 0; i < ncolors; i++ )
		xc[i].pixel = i;
	    XQueryColors(XWinSystem::display, cmap, xc, ncolors);
	    for ( i = 0; i < ncolors; i++ )
		if ( ! color_table[i].is_valid ) {
		    color_table[i].client = color_table[i].server =
			RGB(xc[i].red >> 8, xc[i].green >> 8, xc[i].blue >> 8);
		    color_table[i].is_valid = TRUE;
		}
	    local_colormap = TRUE;
	    delete xc;
	}
    }

    if ( ! found )
    {
	// We could neither find an exact match nor allocate a new color cell.
	// Now look for the nearest real RGB value in "color_table".
	u_long mindist = 256*256*3;
	for ( i = 0; i < ncolors; i++ ) {
	    if ( color_table[i].is_valid ) {
		RGB d = *rgb - color_table[i].server;
		u_long dist = (d.red*d.red + d.green*d.green + d.blue*d.blue);
		if ( dist < mindist ) {
		    mindist = dist;
		    pixel = i;
		    found = TRUE;
		}
	    }
	}
	if(!found) 
	    Warning ("DevRGB2Index", "cannot allocate color %d,%d,%d; using %d,%d,%d\n",
		rgb->red, rgb->green, rgb->blue,
		color_table[pixel].server.red,
		color_table[pixel].server.green,
		color_table[pixel].server.blue);
    }

    // At this point we must have found a "pixel".
    // In the worst case it is either black or white.
    color_table[pixel].ref_cnt++;
    if (result)
	*result = color_table[pixel].server;

    return pixel;
}

void
XColorIndexMapper::DevIndex2RGB(u_long ix, RGB *result)
{
    if (result)
	*result = color_table[ix].server;
}

//---- XColorDirectMapper ------------------------------------------------

static int
findshift(u_long mask)
{
    register int i;
    for ( i = 0; i < 32; mask >>= 1, i++ )
	if ( mask & 1 )
	    return i;
    return 0;
}

static int
findmult(u_long mask, int shift)
{
    register int i;
    mask >>= shift;
    for ( i = 0; mask != 0; i++ )
	mask >>= 1;
    return (8 - i);
}

XColorDirectMapper::XColorDirectMapper(Visual *vis)
    : WindowColorMap((bool)FALSE)
{
    red_mask   = vis->red_mask;
    green_mask = vis->green_mask;
    blue_mask  = vis->blue_mask;
    red_shift   = findshift(red_mask);
    green_shift = findshift(green_mask);
    blue_shift  = findshift(blue_mask);
    red_mult   = findmult(red_mask, red_shift);
    green_mult = findmult(green_mask, green_shift);
    blue_mult  = findmult(blue_mask, blue_shift);
}

u_long
XColorDirectMapper::DevRGB2Index(RGB *rgb, RGB *result)
{
    u_long ix =
	(((u_long(rgb->red) >> red_mult) << red_shift) & red_mask)         |
	(((u_long(rgb->green) >> green_mult) << green_shift) & green_mask) |
	(((u_long(rgb->blue) >> blue_mult) << blue_shift) & blue_mask);
    if (result)
	DevIndex2RGB(ix, result);
    return ix;
}

void
XColorDirectMapper::DevIndex2RGB(u_long ix, RGB *result)
{
    if (result) {
	result->red   = short(((ix & red_mask) >> red_shift) << red_mult);
	result->green = short(((ix & green_mask) >> green_shift) << green_mult);
	result->blue  = short(((ix & blue_mask) >> blue_shift) << blue_mult);
    }
}
