#ifdef __GNUG__
#pragma implementation
#endif

#include "Region.h"
#include "Class.h"
#include "ETString.h"

//---- XXSimpleRefCounted --------------------------------------------------------------
#include "Error.h"

class XXSimpleRefCounted {
public:
    XXSimpleRefCounted();
    XXSimpleRefCounted(long cnt);
    XXSimpleRefCounted(const XXSimpleRefCounted &copy);

    void Ref();
    void Unref();
    #define SafeUnref(p) p->Unref(), p= 0
    long GetRefCount();

protected:
    virtual ~XXSimpleRefCounted();

private:
    long fRefCount;
};

inline XXSimpleRefCounted::XXSimpleRefCounted() : fRefCount(1)
    { }
inline XXSimpleRefCounted::XXSimpleRefCounted(long cnt) : fRefCount(cnt)
    { }
inline void XXSimpleRefCounted::Ref()
    { if (this) fRefCount++; }
inline void XXSimpleRefCounted::Unref()
    { if (this && --fRefCount <= 0) delete this; }
inline long XXSimpleRefCounted::GetRefCount()
    { return fRefCount; }

XXSimpleRefCounted::XXSimpleRefCounted(const XXSimpleRefCounted &copy)
{
    Fatal("XXSimpleRefCounted::XXSimpleRefCounted(const XXSimpleRefCounted &copy)", "");
}

XXSimpleRefCounted::~XXSimpleRefCounted()
{
    fRefCount= 0;
}

//---- RegionImpl --------------------------------------------------------------

class RegionImpl : public XXSimpleRefCounted {
public:
    RegionImpl() : fBuf(0), fCapacity(0)
	{ }
    RegionImpl(const Coord *s, long len);
    RegionImpl(const Rectangle &r);
    RegionImpl(const Region &r);
    ~RegionImpl();

    long GetRects(Rectangle *rp) const;
    bool IsEqual(RegionImpl*) const;
    bool ContainsPoint(const Point &p) const;

    OStream &PrintOn(OStream &os) const;
    IStream &ReadFrom(IStream &is);

protected:
    void CalcBBox();

protected:
friend Region;
    Rectangle fBBox;
    long fCapacity;
    Coord *fBuf, fRectBuf[7];
};

RegionImpl::RegionImpl(const Rectangle &r)
    : fBuf(fRectBuf), fCapacity(7), fBBox(r)
{
    register Coord *s= fBuf;
    *s++= 3; *s++= r.origin.y; *s++= r.origin.x; *s++= r.origin.x+r.extent.x;
    *s++= 1; *s++= r.origin.y+r.extent.y;
    *s= 0;
}

RegionImpl::RegionImpl(const Coord *s, long len)
{
    if (s && len > 0) {
	fCapacity= len;
	if (len <= 7)
	    fBuf= fRectBuf;
	else
	    fBuf= new Coord[len];
	MemCpy(fBuf, s, len*sizeof(Coord));
	CalcBBox();
    } else {
	fCapacity= 0;
	fBuf= 0;
    }
}

RegionImpl::~RegionImpl()
{
    if (fBuf) {
	if (fRectBuf != fBuf)
	    delete fBuf;
	fBuf= 0;
	fCapacity= 0;
    }
}

long RegionImpl::GetRects(Rectangle *rp) const
{
    register Coord *s= fBuf;
    long i= 0;

    if (s) {
	register long next;
	Rectangle r;
	while (next= *s++) {
	    r.origin.y= *s++;
	    r.extent.y= s[next] - r.origin.y;
	    while (next > 1) {
		r.origin.x= *s++;
		r.extent.x= *s++-r.origin.x;
		next-= 2;
		if (r.IsNotEmpty())
		    rp[i++]= r;
	    }
	}
    }
    return i;
}

void RegionImpl::CalcBBox()
{
    register Coord *s= fBuf;
    register long next;
    Rectangle r;

    fBBox= gRect0;
    if (s) {
	while (next= *s++) {
	    r.origin.y= *s++;
	    r.extent.y= s[next] - r.origin.y;
	    while (next > 1) {
		r.origin.x= *s++;
		r.extent.x= *s++-r.origin.x;
		next-= 2;
		fBBox.Merge(r);
	    }
	}
    }
}

bool RegionImpl::ContainsPoint(const Point &p) const
{
    register Coord *bp= fBuf;
    register long s, e, next;

    if (fBBox.ContainsPoint(p)) {
	while (next= *bp++) {
	    s= *bp++;
	    if (p.y >= s && p.y < bp[next]) {
		while (next > 1) {
		    s= *bp++;
		    e= *bp++;
		    if (p.x >= s && p.x < e)
			return TRUE;
		    next-= 2;
		}
	    } else
		bp+= next-1;
	}
    }
    return FALSE;
}

OStream &RegionImpl::PrintOn(OStream &os) const
{
    os << fCapacity SP;
    for (long i= 0; i < fCapacity; i++) {
	os << fBuf[i] SP;
	if (((i+1) % 10) == 0)
	    os << '\n';
    }
    return os NL;
}

IStream &RegionImpl::ReadFrom(IStream &is)
{
    SafeDelete(fBuf);
    is >> fCapacity;
    if (fCapacity > 0) {
	fBuf= new Coord[fCapacity];
	for (long i= 0; i < fCapacity; i++)
	    is >> fBuf[i];
	CalcBBox();
    } else
	fBuf= 0;
    return is;
}

bool RegionImpl::IsEqual(RegionImpl *r2) const
{
    if (fCapacity != r2->fCapacity)
	return FALSE;
    if (fBuf == 0 && r2->fBuf == 0)
	return TRUE;
    if (fBuf == 0 || r2->fBuf == 0)
	return FALSE;
    return memcmp(fBuf, r2->fBuf, fCapacity*sizeof(Coord)) == 0;
}

//---- Region ------------------------------------------------------------------

SimpleMetaImpl(Region)
{
    len= len;
    if (! isptr)
	sprintf(buf, "%d", ((Region*) addr)->Length());
}

Region::Region()
{
    fImpl= new RegionImpl;
}

Region::Region(const Coord *s, long len)
{
    fImpl= new RegionImpl(s, len);
}

Region::Region(const Rectangle &r)
{
    fImpl= new RegionImpl(r);
}

Region::Region(const Region &r)
{
    if (fImpl= r.fImpl)
	fImpl->Ref();
}

Region::~Region()
{
    SafeUnref(fImpl);
}

long Region::GetRects(Rectangle *rp) const
{
    return fImpl->GetRects(rp);
}

Region &Region::operator=(const Region &r)
{
    if (fImpl != r.fImpl) {
	r.fImpl->Ref();
	fImpl->Unref();
	fImpl= r.fImpl;
    }
    return *this;
}

bool Region::ContainsPoint(const Point &p) const
{
    return fImpl->ContainsPoint(p);
}

OStream &Region::PrintOn(OStream &os) const
{
    return fImpl->PrintOn(os);
}

IStream &Region::ReadFrom(IStream &is)
{
    return fImpl->ReadFrom(is);
}

const Rectangle &Region::BBox() const
{
    return fImpl->fBBox;
}

long Region::Length() const
{
    return fImpl->fCapacity;
}

Coord *Region::GetBuf() const
{
    return fImpl->fBuf;
}

bool Region::IsEmpty() const
{
    return fImpl->fBBox.IsEmpty();
}

bool Region::IsNotEmpty() const
{
    return fImpl->fBBox.IsNotEmpty();
}

bool Region::IsEqual(const Region &other) const
{
    if (fImpl == other.fImpl)
	return TRUE;
    return fImpl->IsEqual(other.fImpl);
}

Region RegionOp(Region::EOpcode code, const Region &r1, const Region &r2)
{
    switch (code) {
    case Region::eCopy:
	return r2;
    case Region::eUnion:
	if (r1.IsEmpty() && r2.IsEmpty())
	    return Region();
	if (r1.IsEmpty())
	    return r2;
	if (r2.IsEmpty())
	    return r1;
	break;
    case Region::eDifference:
	if (r1.IsEmpty())
	    return Region();
	if (r2.IsEmpty())
	    return r1;
	break;
    case Region::eRevDifference:
	if (r2.IsEmpty())
	    return Region();
	if (r1.IsEmpty())
	    return r2;
	break;
    case Region::eIntersection:
	if (r1.IsEmpty() || r2.IsEmpty())
	    return Region();
	break;
    }

    Coord *shape1= r1.GetBuf(), *shape2= r2.GetBuf();
    Coord *buf= new Coord[r1.Length()*r2.Length()];
    register Coord *ss= buf, *x1, *x2;
    Coord *fixup, *oldss= 0;
    long y, len, oldlen= -1, l1= 0, l2= 0, tl1, tl2, test;

    for (;;) {
	tl1= *shape1;
	tl2= *shape2;

	if (tl1 == 0 && tl2 == 0)
	    break;

	if (tl1 == 0)
	    test= 1;
	else if (tl2 == 0)
	    test= -1;
	else
	    test= shape1[1]-shape2[1];

	if (test <= 0) {
	    y= shape1[1];
	    x1= &shape1[2];
	    shape1+= tl1+1;
	    l1= tl1-1;
	}
	if (test >= 0) {
	    y= shape2[1];
	    x2= &shape2[2];
	    shape2+= tl2+1;
	    l2= tl2-1;
	}

	fixup= ss;
	*ss++= 1;
	*ss++= y;

	if (l1 == 0 && l2 == 0) {
	    ;
	} else if (l1 == 0) {
	    if (code == Region::eUnion || code == Region::eRevDifference) {
		MemCpy(ss, x2, l2*sizeof(Coord));
		ss+= l2;
	    }
	} else if (l2 == 0) {
	    if (code == Region::eUnion || code == Region::eDifference) {
		MemCpy(ss, x1, l1*sizeof(Coord));
		ss+= l1;
	    }
	} else {
	    register Coord *p1= x1, *p2= x2;
	    register long xl1= l1, xl2= l2;
	    register long x, xold= 0, xflag= 0;

	    while (xl1 > 0 && xl2 > 0) {
		test= *p1-*p2;
		if (test <= 0) {
		    x= *p1++;
		    xflag ^= 1;
		    xl1--;
		}
		if (test >= 0) {
		    x= *p2++;
		    xflag ^= 2;
		    xl2--;
		}
		if (xflag == code || xold == code)
		    *ss++= x;
		xold= xflag;
	    }

	    if (code == Region::eUnion || code == Region::eDifference) {
		while (xl1-- > 0)
		    *ss++= *p1++;
	    }
	    if (code == Region::eUnion || code == Region::eRevDifference) {
		while (xl2-- > 0)
		    *ss++= *p2++;
	    }
	}
	len= ss-fixup-2;
	fixup[0]= len+1;

	if (len > 0 && len == oldlen && memcmp(&oldss[2], &fixup[2], len*sizeof(Coord)) == 0) {
	    ss= fixup;
	} else {
	    oldss= fixup;
	    oldlen= len;
	}
    }
    *ss++= 0;

    return Region(buf, ss-buf);
}

Region SymDifference(const Region &r1, const Region &r2)
{
    Region in(Intersection(r1, r2));
    return Union(Difference(r1, in), Difference(r2, in));
}

//---- RegionRectIter ----------------------------------------------------------

RegionRectIter::RegionRectIter(const Region &r)
    : fRegion(r), fBufPtr(r.GetBuf()), fNext(0)
{
}

bool RegionRectIter::operator()(Rectangle &r)
{
    while (fBufPtr) {
	if (fNext > 1) {
	    r.origin.x= *fBufPtr++;
	    r.extent.x= *fBufPtr++ - r.origin.x;
	    fNext-= 2;
	    return TRUE;
	}
	fNext= *fBufPtr++;
	if (fNext == 0)
	    return FALSE;
	r.origin.y= *fBufPtr++;
	r.extent.y= fBufPtr[fNext]-r.origin.y;
    }
    return FALSE;
}

//---- OldRegion ---------------------------------------------------------------

OldRegion::OldRegion()
{
    size= 0;
    rl= 0;
    bbox= gRect0;
    
    rl= (Rectangle*)Storage::Alloc(sizeof(Rectangle)*25);
}

OldRegion::OldRegion(Rectangle &r)
{
    size= 1;
    rl= (Rectangle*) Storage::Alloc(sizeof(Rectangle)*size);
    rl[0]= r;
    bbox= r;
}

OldRegion::~OldRegion()
{
    size= 0;
    SafeDelete(rl);
}

Rectangle *OldRegion::GetAt(int ix)
{
    if (ix >= 0 && ix < size)
	return &rl[ix];
    return 0;
}

void OldRegion::operator=(OldRegion &r)
{
    SafeDelete(rl);
    size= r.size;
    bbox= r.bbox;
    rl= (Rectangle*) Storage::Alloc(sizeof(Rectangle)*size);
    memcpy(rl, r.rl, sizeof(Rectangle)*size);
}

void OldRegion::Merge(Rectangle r)
{
    if (size == 0) {
	rl[size++]= bbox= r;
	return;
    }
    if (r.ContainsRect(bbox)) {
	size= 0;
	rl[size++]= bbox= r;
	return;
    }
    bbox.Merge(r);
    for (int i= 0; i < size; i++) {
	if (rl[i].ContainsRect(r))
	    return;
	if (r.ContainsRect(rl[i])) {
	    rl[i]= r;
	    return;
	} 
	if (r.Intersects(rl[i])) {
	    rl[i].Merge(r);
	    return;
	}
    }
    if (size >= 20)
	rl[size-1].Merge(r);
    else
	rl[size++]= r;
}

