#ifdef __GNUG__
#pragma implementation
#endif

#include "Storage.h"

#include "CLib.h"
#include "ETString.h"
#include "Error.h"
#include "Object.h"
#include "ObjectTable.h"
#include "ETMath.h"


//#define MEM_DEBUG
#define MEM_MAGIC      0x55
#define MEM_MAGIC_LEN  8

//#define MEM_STAT
//#define MEM_CHECKOBJECTPOINTERS

#ifdef MEM_CHECKOBJECTPOINTERS
#   define CheckObjPtr(name, p) ObjectTable::CheckPtrAndWarn(name, p)
#else
#   define CheckObjPtr(name, p) p
#endif

const int cMemPoolSize  =   100,
	  cObjMaxSize   =   1024;       

u_long Storage_lastobj, Storage_lastobjend;
bool Storage_proto;

#ifdef MEM_DEBUG
static size_t maxA;
#endif

static bool died;
static FreeHookFun storageFreeHook;
static void *storageFreeHookData;

//#define ET_STORAGE__
#ifdef	ET_STORAGE__

//---- Pool for a specific size ------------------------------------------------

struct FreeNode {
    class FreeNode *next;
};

class Chunk {
public:
    struct Chunk *next;
    char *storage;
public:
    Chunk(Chunk *n, size_t size)
	{ next= n; storage= new char[size]; }
    ~Chunk()
	{ SafeDelete(storage); }
};

class MemPool {
public:
friend class Storage;
    struct Chunk *chunks;
    struct FreeNode *freeList;
    int lastIndex, chunkSize, allocated, recycled, freed;
    size_t objSize;
public:
    MemPool(size_t objSize);
    ~MemPool();
    void *Alloc();
    void Free(void *);
    int nchunks();
};

static MemPool *pools[cObjMaxSize];

#ifdef MEM_STAT

    static bool gMemStatistics;
    static int ASizes[cObjMaxSize];
    static int FSizes[cObjMaxSize];
    static int totalA, totalF;
    static void *ttt[1000];
    static int tt, mem_size= -1, mem_ix= -1;
    
    void EnterStat(size_t size, void *p)
    {
	if (gMemStatistics && size == mem_size) {
	    if (tt == mem_ix)
		CLib::Abort();
	    ttt[tt++]= p;
	}
	maxA= Math::Max(maxA, size);
	if (size >= cObjMaxSize)
	    ASizes[cObjMaxSize-1]++;
	else
	    ASizes[size]++;
	totalA+= size;
    }
    
    void RemoveStat(size_t size, void *vp)
    {
	if (gMemStatistics && size == mem_size) {
	    for (int i= 0; i < tt; i++)
		if (ttt[i] == vp) {
		    ttt[i]= 0;
		    break;
		}
	}
	if (size >= cObjMaxSize)
	    FSizes[cObjMaxSize-1]++;
	else
	    FSizes[size]++;
	totalF+= size;
    }
    
#else

#   define EnterStat(s, p) maxA= Math::Max(maxA, s);
#   define RemoveStat(s, p)

#endif

static bool FreeFromPool(void *vp)
{
    register int i;
    register Chunk *p;
    register MemPool *mp;

    for (i= 0; i < cObjMaxSize; i++) {
	if (mp= pools[i]) {
	    for (p= mp->chunks; p; p= p->next) {
		if (vp >= (void*)p->storage
				    && vp < (void*)(p->storage + mp->chunkSize)) {
		    mp->Free(vp);
		    return TRUE;
		}
	    }
	}
    }
    return FALSE;
}

#define storage_size(ptr) ((size_t)(((int*)ptr)[-1]))

//---- Pool for a specific size ------------------------------------------------

MemPool::MemPool(size_t osize)
{
    objSize= osize;
    chunkSize= cMemPoolSize * objSize;
    chunks= 0;
    lastIndex= cMemPoolSize; // force allocation of a new chunk
    freeList= 0;
    allocated= recycled= freed= 0;
}

MemPool::~MemPool()
{
    register Chunk *p, *lp;

    for (p= chunks; p; p= lp) {
	lp= p->next;
	delete p;
    }
}

int MemPool::nchunks()
{
    register int n;
    register Chunk *p;
    
    for (n= 0, p= chunks; p; p= p->next, n++)
	;
    return n;
}

void *MemPool::Alloc()
{
    allocated++;
    if (freeList) {
	recycled++;
	FreeNode *tmp= freeList;
	freeList= freeList->next;
	return (void*)tmp;
    }
    if (lastIndex >= cMemPoolSize) { // need new chunk
	chunks= new Chunk(chunks, chunkSize);
	lastIndex= 0;
    }
    return (void*) &chunks->storage[lastIndex++ * objSize];
}

void MemPool::Free(void *op)
{
    if (storageFreeHook)
	storageFreeHook(storageFreeHookData, op, objSize);

    for (Chunk *p= chunks; p; p= p->next) {
	if (op >= (void*)p->storage && op <  (void*)(p->storage + chunkSize)) {
	    memset(op, 0, objSize);
	    FreeNode *fn= (FreeNode*)op;
	    fn->next= freeList;
	    freeList= fn;
	    freed++;
	    return;
	}
    }
    fprintf(stderr, "Fatal in MemPool::Free: object to delete not found in pool %d\n", objSize);
    fflush(stderr);
    CLib::Abort();
}

//---- Storage -----------------------------------------------------------------

void* operator new(size_t size)
{
    register char *sp;

    if (size < 0) {
	fprintf(stderr, "Fatal in operator new: size < 0\n");
	fflush(stderr);
	CLib::Abort();
    }
#ifdef MEM_DEBUG
# ifdef MEM_MAGIC
    sp= (char*) CLib::CAlloc(sizeof(int) + size + MEM_MAGIC_LEN, sizeof(char));
# else
    sp= (char*) CLib::CAlloc(sizeof(int) + size, sizeof(char));
# endif
    *((int*)sp) = size;
# ifdef MEM_MAGIC
    memset(sp+size+sizeof(int), MEM_MAGIC, MEM_MAGIC_LEN);
# endif
    EnterStat(size, sp + sizeof(int));
    return CheckObjPtr("operator new", sp + sizeof(int));
#else
    sp= (char*) CLib::CAlloc(size, sizeof(char));
    return CheckObjPtr("operator new", sp);
#endif
}

void operator delete(void* ptr)
{
    Storage_lastobj= Storage_lastobjend= 0;
    Storage_proto= FALSE;
#ifdef MEM_DEBUG
    if (ptr) {
	(void) CheckObjPtr("operator delete", ptr);
	if (!died && FreeFromPool(ptr))
	    return;
	int size= storage_size((char*)ptr);
	if (size < 0 || size > maxA) {
	    fprintf(stderr, "Fatal in operator delete: unreasonable size (%d)\n", size);
	    fflush(stderr);
	    CLib::Abort();
	}
	RemoveStat(size, ptr);
	if (storageFreeHook) 
	    storageFreeHook(storageFreeHookData, ptr, size);
	ptr= (char*)ptr - sizeof(int);
	size+=sizeof(int);
# ifdef MEM_MAGIC
	register int i;
	register char *tmp_ptr = (char*)ptr + size;
	for (i = 0; i < MEM_MAGIC_LEN; i++)
	    if (tmp_ptr[i] != MEM_MAGIC) {
		fprintf(stderr, "operator delete: storage area overwritten\n");
		fflush(stderr);
		CLib::Abort();
	    }
	size += MEM_MAGIC_LEN;
# endif
	memset(ptr, 0, size);
	CLib::Free((char*) ptr);
    }
#else
    if (ptr) {
	(void) CheckObjPtr("operator delete", ptr);
	if (!died && FreeFromPool(ptr))
	    return;
	if (storageFreeHook) 
	    storageFreeHook(storageFreeHookData, ptr, 0);
	CLib::Free((char*) ptr);
    }
#endif
}

void *Storage::Alloc(size_t sz)
{
    return new char[sz];
}

void Storage::Free(void *vp)
{
    delete vp;
}

void *Storage::ReAlloc(void *vp, size_t size)
{
    if (vp == 0)
	return new char[size];
#ifdef MEM_DEBUG
    int oldsize= storage_size((char*)vp);
    if (size > oldsize) {
	int *sp, sz= size;
	char *ptr= (char*)vp, *cp;
    
	RemoveStat(oldsize, vp);
	
#ifdef MEM_MAGIC
	sz+= MEM_MAGIC_LEN;
#endif
	sp= (int*) CLib::ReAlloc((char*) (ptr - sizeof(int)), sz+sizeof(int));
	cp= (char*) &sp[1];
	EnterStat(size, &sp[1]);
	
#ifdef MEM_MAGIC
	register int i;
	register char *tmp_ptr = &cp[oldsize];
	for (i = 0; i < MEM_MAGIC_LEN; i++)
	    if (tmp_ptr[i] != MEM_MAGIC) {
		fprintf(stderr, "Storage::Realloc: storage area overwritten\n");
		fflush(stderr);
		CLib::Abort();
	    }
	memset((char*)sp+size+sizeof(int), MEM_MAGIC, MEM_MAGIC_LEN);
#endif
	memset(&cp[oldsize], 0, size-oldsize);
	sp[0]= size;
	return CheckObjPtr("Storage::ReAlloc", cp);
    }
    return vp;
#else
    vp= (void*) CLib::ReAlloc((char*)vp, size);
    return CheckObjPtr("Storage::ReAlloc", vp);
#endif
}

void *Storage::ChunkAlloc(size_t sz)
{
    if (died || sz >= cObjMaxSize)
	return (void*) new char[sz];
    MemPool *mp= pools[sz];
    if (mp == 0)
	mp= pools[sz]= new MemPool(sz);
    return CheckObjPtr("Storage::ChunkAlloc", mp->Alloc());
}

void Storage::ChunkFree(void *vp)
{
    register int i;
    register Chunk *p;
    register MemPool *mp;

    for (i= 0; i < cObjMaxSize; i++) {
	if (mp= pools[i]) {
	    for (p= mp->chunks; p; p= p->next) {
		if (vp >= (void*)p->storage
				    && vp < (void*)(p->storage + mp->chunkSize)) {
		    mp->Free(vp);
		    return;
		}
	    }
	}
    }
}

void *Storage::ObjectChunkAlloc(size_t sz)
{
    if (died || sz >= cObjMaxSize)
	return (void*) new char[sz];
    MemPool *mp= pools[sz];
    if (mp == 0) 
	mp= pools[sz]= new MemPool(sz);
    return CheckObjPtr("Storage::ObjectChunkAlloc", mp->Alloc());
}

void Storage::ObjectChunkFree(void *vp)
{
    delete vp;
}

void *Storage::ObjectAlloc(size_t sz)
{
    Storage_proto= FALSE;
    Storage_lastobj= (u_long) new char[sz];
    Storage_lastobjend= Storage_lastobj + sz;
    return (void*) Storage_lastobj;
}

void Storage::ObjectFree(void *vp)
{
    delete vp;
}

void Storage::SetFreeHook(FreeHookFun fh, void *data)
{
    storageFreeHook= fh;
    storageFreeHookData= data;
}

void Storage::Statistics()
{
#ifdef MEM_DEBUG
#ifdef MEM_STAT
    MemPool *fsp;
	int	sum = 0;

//    if (!gMemStatistics)
//	return;
	
    if (totalA == totalF) {
	fprintf(stdout, "no garbage left (total: %d)\n", totalA);
	return;
    }
    
    fprintf(stdout, "\n");
    fprintf(stdout, "Mem statistics\n");
    fprintf(stdout, "%7s%9s%9s%9s%9s%9s%9s%9s",
	"size", "alloc", "free", "diff", "alloc", "recycl", "freed", "chunks\n");
    fprintf(stdout, "=====================================================================\n");
    

    for (int i= 0; i < cObjMaxSize; i++) 
    	{
		fsp= pools[i];
		int diff = ASizes[i]-FSizes[i];
		if (diff || fsp)
		    fprintf(stdout, "%7d%9d%9d%9d", i, ASizes[i], FSizes[i],
								diff);
		if (fsp)
		    fprintf(stdout, "%9d%9d%9d%9d", fsp->allocated, 
					    fsp->recycled, fsp->freed, fsp->nchunks());
		
		if (diff || fsp)
			{
			sum += i * diff;
			if (fsp)
				fprintf(stdout, "%9d%9d", i * diff, sum);
			else
				fprintf(stdout, "%45d%9d", i * diff, sum);
			
		    fprintf(stdout, "\n");
		    }
    	}

    if (totalA != totalF) {
	fprintf(stdout, "---------------------------------------------------------------------\n");
	fprintf(stdout, "total: %9d%9d%9d\n", totalA, totalF, totalA-totalF);
    }

    if (mem_size != -1) {
	fprintf(stdout, "---------------------------------------------------------------------\n");
	for (i= 0; i < tt; i++)
	    if (ttt[i])
		fprintf(stdout, "block %d of size %d not freed\n", i, mem_size);
    }
    fprintf(stdout, "=====================================================================\n");
    fprintf(stdout, "\n");
    fflush(stdout);
#endif
#endif
}

void Storage::EnableStatistics(int size, int ix)
{
#ifdef MEM_STAT
    mem_size= size;
    mem_ix= ix;
    gMemStatistics= TRUE;
#endif
}

void Storage::FreeAll()
{
    died= TRUE;
    for (int i= 0; i < cObjMaxSize; i++)
	SafeDelete(pools[i]);

    Statistics();
}

#else	// ET_STORAGE

#include <stdlib.h>
void* operator new(size_t sz)
{
	return (void*)CLib::CAlloc(sz,sizeof(char));
}

void *operator new(size_t size, void *vp) //ruedi
{
    return ::operator new(size);
}

#ifdef OS_WINDOWS  //ruedi: for some reason GCC doesn't like that!?!
    void *operator new[](size_t size)	//ruedi
    {
	return ::operator new(size);
    }
#endif
void operator delete(void* ptr)
{
	CLib::Free((char*)ptr);
}

void *Storage::Alloc(size_t sz)
{
	return (void*)CLib::CAlloc(sz,sizeof(char));
}
void Storage::Free(void *vp)
{
	CLib::Free((char*)vp);
}

void *Storage::ChunkAlloc(size_t sz)
{
	return (void*)CLib::CAlloc(sz,sizeof(char));
}
void Storage::ChunkFree(void *vp)
{
	CLib::Free((char*)vp);
}

void *Storage::ObjectChunkAlloc(size_t sz)
{
	return (void*)CLib::CAlloc(sz,sizeof(char));
}
void Storage::ObjectChunkFree(void *vp)
{
	CLib::Free((char*)vp);
}

void *Storage::ObjectAlloc(size_t sz)
{
	return (void*)CLib::CAlloc(sz,sizeof(char));
}
void Storage::ObjectFree(void *vp)
{
	CLib::Free((char*)vp);
}

void *Storage::ReAlloc(void *vp, size_t sz)
{
	if(vp)
		return (void*)CLib::ReAlloc((char*)vp, (int)sz);
	else
		return (void*)CLib::CAlloc(sz,sizeof(char));
}

void Storage::Statistics()
{
}
void Storage::SetFreeHook(FreeHookFun, void *data)
{
}
void Storage::FreeAll()
{
}
void Storage::EnableStatistics(int size, int ix)
{
}

#endif
